import enum
import os
import sys
from dataclasses import dataclass
from typing import List, Optional, Tuple, Union


class Endianness(str, enum.Enum):
    """
    An enum of endian values: big, little, not_applicable.
    """

    big = "big"
    little = "little"
    not_applicable = "not_applicable"


class ObjectArrayTypeDisabled(ValueError):
    pass


class Kind(str, enum.Enum):
    """
    See https://numpy.org/devdocs/reference/arrays.interface.html#object.__array_interface__

    The term "kind" comes from the numpy API as well.

    Note: At import time, the environment variable ``TILED_ALLOW_OBJECT_ARRAYS``
    is checked. If it is set to anything other than ``"0"``, then this
    Enum gets an additional member::

        object = "O"

    to support numpy 'object'-type arrays which hold generic Python objects.
    Numpy 'object'-type arrays are not enabled by default because their binary
    representation is not interpretable by clients other than Python.  It is
    recommended to convert your data to a non-object type if possible so that it
    can be read by non-Python clients.
    """

    bit_field = "t"
    boolean = "b"
    integer = "i"
    unsigned_integer = "u"
    floating_point = "f"
    complex_floating_point = "c"
    timedelta = "m"
    datetime = "M"
    string = "S"  # fixed-length sequence of char
    unicode = "U"  # fixed-length sequence of Py_UNICODE
    other = "V"  # "V" is for "void" -- generic fixed-size chunk of memory

    # By default, do not tolerate numpy objectg arrays
    if os.getenv("TILED_ALLOW_OBJECT_ARRAYS", "0") != "0":
        object = "O"  # Object (i.e. the memory contains a pointer to PyObject)

    @classmethod
    def _missing_(cls, key):
        if key == "O":
            raise ObjectArrayTypeDisabled(
                "Numpy 'object'-type arrays are not enabled by default "
                "because their binary representation is not interpretable "
                "by clients other than Python. "
                "It is recommended to convert your data to a non-object type "
                "if possible so that it can be read by non-Python clients. "
                "If this is not possible, you may enable 'object'-type arrays "
                "by setting the environment variable TILED_ALLOW_OBJECT_ARRAYS=1 "
                "in the server."
            )


@dataclass
class BuiltinDtype:
    endianness: Endianness
    kind: Kind
    itemsize: int

    __endianness_map = {
        ">": "big",
        "<": "little",
        "=": sys.byteorder,
        "|": "not_applicable",
    }

    __endianness_reverse_map = {"big": ">", "little": "<", "not_applicable": "|"}

    @classmethod
    def from_numpy_dtype(cls, dtype):
        return cls(
            endianness=cls.__endianness_map[dtype.byteorder],
            kind=Kind(dtype.kind),
            itemsize=dtype.itemsize,
        )

    def to_numpy_dtype(self):
        import numpy

        return numpy.dtype(self.to_numpy_str())

    def to_numpy_str(self):
        endianness = self.__endianness_reverse_map[self.endianness]
        # dtype.itemsize always reports bytes.  The format string from the
        # numeric types the string format is: {type_code}{byte_count} so we can
        # directly use the item size.
        #
        # for unicode the pattern is 'U{char_count}', however
        # which numpy always represents as 4 byte UCS4 encoding
        # (because variable width encodings do not fit with fixed-stride arrays)
        # so the reported itemsize is 4x the char count.  To get back to the string
        # we need to divide by 4.
        size = self.itemsize if self.kind != Kind.unicode else self.itemsize // 4
        return f"{endianness}{self.kind.value}{size}"

    @classmethod
    def from_json(cls, structure):
        return cls(
            kind=Kind(structure["kind"]),
            itemsize=structure["itemsize"],
            endianness=Endianness(structure["endianness"]),
        )


@dataclass
class Field:
    name: str
    dtype: Union[BuiltinDtype, "StructDtype"]
    shape: Optional[Tuple[int, ...]]

    @classmethod
    def from_numpy_descr(cls, field):
        import numpy

        name, *rest = field
        if name == "":
            raise ValueError(
                f"You seem to have gotten descr of a base or subdtype: {field}"
            )
        if len(rest) == 1:
            (f_type,) = rest
            shape = None
        else:
            f_type, shape = rest

        if isinstance(f_type, str):
            FType = BuiltinDtype.from_numpy_dtype(numpy.dtype(f_type))
        else:
            FType = StructDtype.from_numpy_dtype(numpy.dtype(f_type))
        return cls(name=name, dtype=FType, shape=shape)

    def to_numpy_descr(self):
        if isinstance(self.dtype, BuiltinDtype):
            base = [self.name, self.dtype.to_numpy_str()]
        else:
            base = [self.name, self.dtype.to_numpy_descr()]
        if self.shape is None:
            return tuple(base)
        else:
            return tuple(base + [self.shape])

    @classmethod
    def from_json(cls, structure):
        name = structure["name"]
        if "fields" in structure["dtype"]:
            ftype = StructDtype.from_json(structure["dtype"])
        else:
            ftype = BuiltinDtype.from_json(structure["dtype"])
        return cls(name=name, dtype=ftype, shape=structure["shape"])


@dataclass
class StructDtype:
    itemsize: int
    fields: List[Field]

    @classmethod
    def from_numpy_dtype(cls, dtype):
        # subdtypes push extra dimensions into arrays, we should handle these
        # a layer up and report an array with bigger dimensions.
        if dtype.subdtype is not None:
            raise ValueError(f"We do not know how to encode subdtypes: {dtype}")
        # If this is a builtin type, require the use of BuiltinDtype (nee .array.BuiltinDtype)
        if dtype.fields is None:
            raise ValueError(f"You have a base type: {dtype}")
        return cls(
            itemsize=dtype.itemsize,
            fields=[Field.from_numpy_descr(f) for f in dtype.descr],
        )

    def to_numpy_dtype(self):
        import numpy

        return numpy.dtype(self.to_numpy_descr())

    def to_numpy_descr(self):
        return [f.to_numpy_descr() for f in self.fields]

    def max_depth(self):
        return max(
            1 if isinstance(f.dtype, BuiltinDtype) else 1 + f.dtype.max_depth()
            for f in self.fields
        )

    @classmethod
    def from_json(cls, structure):
        return cls(
            itemsize=structure["itemsize"],
            fields=[Field.from_json(f) for f in structure["fields"]],
        )


@dataclass
class ArrayStructure:
    data_type: Union[BuiltinDtype, StructDtype]
    chunks: Tuple[Tuple[int, ...], ...]  # tuple-of-tuples-of-ints like ((3,), (3,))
    shape: Tuple[int, ...]  # tuple of ints like (3, 3)
    dims: Optional[Tuple[str, ...]] = None  # None or tuple of names like ("x", "y")
    resizable: Union[bool, Tuple[bool, ...]] = False

    @classmethod
    def from_json(cls, structure):
        if "fields" in structure["data_type"]:
            data_type = StructDtype.from_json(structure["data_type"])
        else:
            data_type = BuiltinDtype.from_json(structure["data_type"])
        dims = structure["dims"]
        if dims is not None:
            dims = tuple(dims)
        return cls(
            data_type=data_type,
            chunks=tuple(map(tuple, structure["chunks"])),
            shape=tuple(structure["shape"]),
            dims=dims,
            resizable=structure.get("resizable", False),
        )

    @classmethod
    def from_array(cls, array, shape=None, chunks=None, dims=None):
        from dask.array.core import normalize_chunks

        if not hasattr(array, "__array__"):
            # may be a list of something; convert to array
            import numpy

            array = numpy.asanyarray(array)

        # Why would shape ever be different from array.shape, you ask?
        # Some formats (notably Zarr) force shape to be a multiple of
        # a chunk size, such that array.shape may include a margin beyond the
        # actual data.
        if shape is None:
            shape = array.shape
        if chunks is None:
            if hasattr(array, "chunks"):
                chunks = array.chunks  # might be None
            else:
                chunks = None
            if chunks is None:
                chunks = ("auto",) * len(shape)
        normalized_chunks = normalize_chunks(
            chunks,
            shape=shape,
            dtype=array.dtype,
        )
        if array.dtype.fields is not None:
            data_type = StructDtype.from_numpy_dtype(array.dtype)
        else:
            data_type = BuiltinDtype.from_numpy_dtype(array.dtype)
        return ArrayStructure(
            data_type=data_type, shape=shape, chunks=normalized_chunks, dims=dims
        )



import base64
import io

import numpy

from ..media_type_registration import deserialization_registry, serialization_registry
from ..utils import (
    SerializationError,
    UnsupportedShape,
    modules_available,
    safe_json_dump,
)


def as_buffer(array, metadata):
    "Give back a zero-copy memoryview of the array if possible. Otherwise, copy to bytes."
    # The memoryview path fails for datetime type (and possibly some others?)
    # but it generally works for standard types like int, float, bool, str.
    try:
        return memoryview(numpy.ascontiguousarray(array))
    except ValueError:
        return numpy.asarray(array).tobytes()


serialization_registry.register(
    "array",
    "application/octet-stream",
    as_buffer,
)
if modules_available("orjson"):
    serialization_registry.register(
        "array",
        "application/json",
        lambda array, metadata: safe_json_dump(array),
    )


def serialize_csv(array, metadata):
    if array.ndim > 2:
        raise UnsupportedShape(array.shape)
    file = io.StringIO()
    numpy.savetxt(file, array, fmt="%s", delimiter=",")
    return file.getvalue().encode()


serialization_registry.register("array", "text/csv", serialize_csv)
serialization_registry.register("array", "text/x-comma-separated-values", serialize_csv)
serialization_registry.register("array", "text/plain", serialize_csv)
deserialization_registry.register(
    "array",
    "application/octet-stream",
    lambda buffer, dtype, shape: numpy.frombuffer(buffer, dtype=dtype).reshape(shape),
)
if modules_available("PIL"):

    def save_to_buffer_PIL(array, format):
        # The logic of which shapes are support is subtle, and we'll leave the details
        # PIL ("beg forgiveness rather than ask permission"). But we can rule out
        # anything above 3 dimensions as definitely not supported.
        if array.ndim > 3:
            raise UnsupportedShape(array.ndim)
        from PIL import Image

        from .image_serializer_helpers import img_as_ubyte

        # Handle too *few* dimensions here, and let PIL raise if there are too
        # *many* because it depends on the shape (RGB, RGBA, etc.)
        array = numpy.atleast_2d(array).astype(numpy.float32)
        # Auto-scale. TODO Use percentile.
        low = numpy.percentile(array.ravel(), 1)
        high = numpy.percentile(array.ravel(), 99)
        scaled_array = numpy.clip((array - low) / (high - low), 0, 1)
        file = io.BytesIO()
        try:
            prepared_array = img_as_ubyte(scaled_array)
            image = Image.fromarray(prepared_array)
            image.save(file, format=format)
        except (TypeError, ValueError):
            raise SerializationError(
                f"Failed to serialize this array as {format}. "
                f"Shape is {array.shape}, dtype is {array.dtype}."
            )

        return file.getbuffer()

    def array_from_buffer_PIL(buffer, format, dtype, shape):
        from PIL import Image

        file = io.BytesIO(buffer)
        image = Image.open(file, format=format)
        return numpy.asarray(image).asdtype(dtype).reshape(shape)

    serialization_registry.register(
        "array", "image/png", lambda array, metadata: save_to_buffer_PIL(array, "png")
    )
    deserialization_registry.register(
        "array",
        "image/png",
        lambda buffer, dtype, shape: array_from_buffer_PIL(buffer, "png", dtype, shape),
    )
if modules_available("tifffile"):

    def array_from_buffer_tifffile(buffer, dtype, shape):
        from tifffile import imread

        return imread(buffer).astype(dtype).reshape(shape)

    def save_to_buffer_tifffile(array, metadata):
        from tifffile import imwrite

        # Handle too *few* dimensions here, and let tifffile raise if there are too
        # *many* because it depends on the shape (RGB, RGBA, etc.)
        normalized_array = numpy.atleast_2d(array)
        # The logic of which shapes are support is subtle, and we'll leave the details
        # tifffile ("beg forgiveness rather than ask permission"). But we can rule out
        # anything above 4 dimensions as definitely not supported.
        if normalized_array.ndim > 4:
            raise UnsupportedShape(array.ndim)
        file = io.BytesIO()
        imwrite(file, normalized_array)
        return file.getbuffer()

    serialization_registry.register("array", "image/tiff", save_to_buffer_tifffile)
    deserialization_registry.register("array", "image/tiff", array_from_buffer_tifffile)


def serialize_html(array, metadata):
    "Try to display as image. Fall back to CSV."
    try:
        png_data = serialization_registry.dispatch("array", "image/png")(
            array, metadata
        )
    except Exception:
        csv_data = serialization_registry.dispatch("array", "text/csv")(array, metadata)
        return "<html>" "<body>" f"{csv_data.decode()!s}" "</body>" "</html>"
    else:
        return (
            "<html>"
            "<body>"
            '<img src="data:image/png;base64,'
            f'{base64.b64encode(png_data).decode()!s}"'
            "/>"
            "</body>"
            "</html>"
        )


serialization_registry.register("array", "text/html", serialize_html)



"""Enrich DataSource-Asset association.

Revision ID: a66028395cab
Revises: 3db11ff95b6c
Create Date: 2024-01-21 15:17:20.571763

"""
import sqlalchemy as sa
from alembic import op

from tiled.catalog.orm import JSONVariant

# revision identifiers, used by Alembic.
revision = "a66028395cab"
down_revision = "3db11ff95b6c"
branch_labels = None
depends_on = None


def upgrade():
    connection = op.get_bind()
    data_sources = sa.Table(
        "data_sources",
        sa.MetaData(),
        sa.Column("id", sa.Integer),
        sa.Column("node_id", sa.Integer),
        sa.Column("mimetype", sa.Unicode),
        sa.Column("structure", JSONVariant),
    )
    data_source_asset_association = sa.Table(
        "data_source_asset_association",
        sa.MetaData(),
        sa.Column("asset_id", sa.Integer),
        sa.Column("data_source_id", sa.Integer),
        sa.Column("parameter", sa.Unicode(255)),
        sa.Column("num", sa.Integer),
    )
    assets = sa.Table(
        "assets",
        sa.MetaData(),
        sa.Column("id", sa.Integer),
        sa.Column("data_uri", sa.Unicode),
    )

    # Rename some MIME types.

    # While Awkward data is typically _transmitted_ in a ZIP archive,
    # it is stored as directory of buffers, with no ZIP involved.
    # Thus using 'application/zip' in the database was a mistake.
    connection.execute(
        data_sources.update()
        .where(data_sources.c.mimetype == "application/zip")
        .values(mimetype="application/x-awkward-buffers")
    )
    # The format is standard parquet. We will use a MIME type
    # parameter to let Tiled know to use the Adapter for sparse
    # data, as opposed to the Adapter for tabular data.
    connection.execute(
        data_sources.update()
        .where(data_sources.c.mimetype == "application/x-parquet-sparse")
        .values(mimetype="application/x-parquet;structure=sparse")
    )

    # Add columns 'parameter' and 'num' to association table.
    op.add_column(
        "data_source_asset_association",
        sa.Column("parameter", sa.Unicode(255), nullable=True),
    )
    op.add_column(
        "data_source_asset_association",
        sa.Column("num", sa.Integer, nullable=True),
    )

    # First populate the columns to bring them into compliance with
    # constraints. Then, apply constraints.

    results = connection.execute(
        sa.select(data_sources.c.id)
        .where(
            sa.not_(
                data_sources.c.mimetype.in_(
                    [
                        "multipart/related;type=image/tiff",
                        "application/x-parquet",
                        "application/x-parquet;structure=sparse",
                    ]
                )
            )
        )
        .select_from(data_sources)
        .join(
            data_source_asset_association,
            data_sources.c.id == data_source_asset_association.c.data_source_id,
        )
        .join(
            assets,
            data_source_asset_association.c.asset_id == assets.c.id,
        )
        .distinct()
    ).fetchall()
    for (data_source_id,) in results:
        connection.execute(
            data_source_asset_association.update()
            .where(data_source_asset_association.c.data_source_id == data_source_id)
            .values(parameter="data_uri")
        )
    results = connection.execute(
        sa.select(data_sources.c.id)
        .where(
            data_sources.c.mimetype.in_(
                [
                    "multipart/related;type=image/tiff",
                    "application/x-parquet",
                    "application/x-parquet;structure=sparse",
                ]
            )
        )
        .select_from(data_sources)
        .join(
            data_source_asset_association,
            data_sources.c.id == data_source_asset_association.c.data_source_id,
        )
        .join(
            assets,
            data_source_asset_association.c.asset_id == assets.c.id,
        )
        .distinct()
    ).fetchall()
    for (data_source_id,) in results:
        connection.execute(
            data_source_asset_association.update()
            .where(data_source_asset_association.c.data_source_id == data_source_id)
            .values(parameter="data_uris")  # plural
        )
        sorted_assoc = connection.execute(
            sa.select(data_source_asset_association.c.data_source_id, assets.c.id)
            .where(data_source_asset_association.c.data_source_id == data_source_id)
            .order_by(assets.c.data_uri)
            .select_from(data_sources)
            .join(
                data_source_asset_association,
                data_sources.c.id == data_source_asset_association.c.data_source_id,
            )
            .join(
                assets,
                data_source_asset_association.c.asset_id == assets.c.id,
            )
        ).fetchall()
        for num, (data_source_id, asset_id) in enumerate(sorted_assoc, start=1):
            connection.execute(
                data_source_asset_association.update()
                .where(data_source_asset_association.c.data_source_id == data_source_id)
                .where(data_source_asset_association.c.asset_id == asset_id)
                .values(num=num)
            )

    # Create unique constraint and triggers.
    if connection.engine.dialect.name == "sqlite":
        # SQLite does not supported adding constraints to an existing table.
        # We invoke its 'copy and move' functionality.
        with op.batch_alter_table("data_source_asset_association") as batch_op:
            # Gotcha: This does not take table_name because it is bound into batch_op.
            batch_op.create_unique_constraint(
                "parameter_num_unique_constraint",
                [
                    "data_source_id",
                    "parameter",
                    "num",
                ],
            )
        # This creates a pair of triggers on the data_source_asset_association
        # table. Each pair include one trigger that runs when NEW.num IS NULL and
        # one trigger than runs when NEW.num IS NOT NULL. Thus, for a given insert,
        # only one of these triggers is run.
        with op.get_context().autocommit_block():
            connection.execute(
                sa.text(
                    """
    CREATE TRIGGER cannot_insert_num_null_if_num_exists
    BEFORE INSERT ON data_source_asset_association
    WHEN NEW.num IS NULL
    BEGIN
        SELECT RAISE(ABORT, 'Can only insert num=NULL if no other row exists for the same parameter')
        WHERE EXISTS
        (
            SELECT 1
            FROM data_source_asset_association
            WHERE parameter = NEW.parameter
            AND data_source_id = NEW.data_source_id
        );
    END"""
                )
            )
            connection.execute(
                sa.text(
                    """
    CREATE TRIGGER cannot_insert_num_int_if_num_null_exists
    BEFORE INSERT ON data_source_asset_association
    WHEN NEW.num IS NOT NULL
    BEGIN
        SELECT RAISE(ABORT, 'Can only insert INTEGER num if no NULL row exists for the same parameter')
        WHERE EXISTS
        (
            SELECT 1
            FROM data_source_asset_association
            WHERE parameter = NEW.parameter
            AND num IS NULL
            AND data_source_id = NEW.data_source_id
        );
    END"""
                )
            )
    else:
        # PostgreSQL
        op.create_unique_constraint(
            "parameter_num_unique_constraint",
            "data_source_asset_association",
            [
                "data_source_id",
                "parameter",
                "num",
            ],
        )
        connection.execute(
            sa.text(
                """
CREATE OR REPLACE FUNCTION raise_if_parameter_exists()
RETURNS TRIGGER AS $$
BEGIN
    IF EXISTS (
        SELECT 1
        FROM data_source_asset_association
        WHERE parameter = NEW.parameter
        AND data_source_id = NEW.data_source_id
    ) THEN
        RAISE EXCEPTION 'Can only insert num=NULL if no other row exists for the same parameter';
    END IF;
    RETURN NEW;
END;
$$ LANGUAGE plpgsql;"""
            )
        )
        connection.execute(
            sa.text(
                """
CREATE TRIGGER cannot_insert_num_null_if_num_exists
BEFORE INSERT ON data_source_asset_association
FOR EACH ROW
WHEN (NEW.num IS NULL)
EXECUTE FUNCTION raise_if_parameter_exists();"""
            )
        )
        connection.execute(
            sa.text(
                """
CREATE OR REPLACE FUNCTION raise_if_null_parameter_exists()
RETURNS TRIGGER AS $$
BEGIN
    IF EXISTS (
        SELECT 1
        FROM data_source_asset_association
        WHERE parameter = NEW.parameter
        AND data_source_id = NEW.data_source_id
        AND num IS NULL
    ) THEN
        RAISE EXCEPTION 'Can only insert INTEGER num if no NULL row exists for the same parameter';
    END IF;
    RETURN NEW;
END;
$$ LANGUAGE plpgsql;"""
            )
        )
        connection.execute(
            sa.text(
                """
CREATE TRIGGER cannot_insert_num_int_if_num_null_exists
BEFORE INSERT ON data_source_asset_association
FOR EACH ROW
WHEN (NEW.num IS NOT NULL)
EXECUTE FUNCTION raise_if_null_parameter_exists();"""
            )
        )


def downgrade():
    # This _could_ be implemented but we will wait for a need since we are
    # still in alpha releases.
    raise NotImplementedError



import io

from ..media_type_registration import deserialization_registry, serialization_registry
from ..utils import modules_available

if modules_available("h5py"):

    def serialize_hdf5(sparse_arr, metadata):
        """
        Place coords and data in HDF5 datasets with those names.
        """
        import h5py

        buffer = io.BytesIO()
        with h5py.File(buffer, mode="w") as file:
            file.create_dataset("data", data=sparse_arr.data)
            file.create_dataset("coords", data=sparse_arr.coords)
            for k, v in metadata.items():
                file.attrs.create(k, v)
        return buffer.getbuffer()

    serialization_registry.register("sparse", "application/x-hdf5", serialize_hdf5)

if modules_available("pandas", "pyarrow"):
    import pandas

    from .table import (
        APACHE_ARROW_FILE_MIME_TYPE,
        XLSX_MIME_TYPE,
        deserialize_arrow,
        serialize_arrow,
        serialize_csv,
        serialize_excel,
        serialize_html,
        serialize_parquet,
    )

    # Support DataFrame formats by first converting to DataFrame.
    # naming columns like dim0, dim1, ..., dimN, data.
    def to_dataframe(sparse_arr):
        d = {f"dim{i}": coords for i, coords in enumerate(sparse_arr.coords)}
        d["data"] = sparse_arr.data
        return pandas.DataFrame(d)

    deserialization_registry.register(
        "sparse", APACHE_ARROW_FILE_MIME_TYPE, deserialize_arrow
    )
    serialization_registry.register(
        "sparse",
        APACHE_ARROW_FILE_MIME_TYPE,
        lambda sparse_arr, metadata: serialize_arrow(
            to_dataframe(sparse_arr), metadata, preserve_index=False
        ),
    )
    serialization_registry.register(
        "sparse",
        "application/x-parquet",
        lambda sparse_arr, metadata: serialize_parquet(
            to_dataframe(sparse_arr), metadata, preserve_index=False
        ),
    )
    serialization_registry.register(
        "sparse",
        "text/csv",
        lambda sparse_arr, metadata: serialize_csv(
            to_dataframe(sparse_arr), metadata, preserve_index=False
        ),
    )
    serialization_registry.register(
        "sparse",
        "text/plain",
        lambda sparse_arr, metadata: serialize_csv(
            to_dataframe(sparse_arr), metadata, preserve_index=False
        ),
    )
    serialization_registry.register(
        "sparse",
        "text/html",
        lambda sparse_arr, metadata: serialize_html(
            to_dataframe(sparse_arr), metadata, preserve_index=False
        ),
    )
    serialization_registry.register(
        "sparse",
        XLSX_MIME_TYPE,
        lambda sparse_arr, metadata: serialize_excel(
            to_dataframe(sparse_arr), metadata, preserve_index=False
        ),
    )
    if modules_available("orjson"):
        import orjson

        def serialize_json(sparse_arr, metadata):
            df = to_dataframe(sparse_arr)
            return orjson.dumps(
                {column: df[column].tolist() for column in df},
            )

        serialization_registry.register(
            "sparse",
            "application/json",
            serialize_json,
        )



from pathlib import Path

import numpy
import psutil
import pytest

from ..catalog import in_memory
from ..catalog.register import register
from ..client import Context, from_context
from ..server.app import build_app, build_app_from_config
from ..server.object_cache import NO_CACHE, ObjectCache, get_object_cache


def test_tallying_hits_and_misses():
    cache = ObjectCache(1e6)
    assert cache.get("a") is None
    assert cache.misses == 1
    assert cache.hits == 0
    assert cache.get("a") is None
    assert cache.misses == 2
    assert cache.hits == 0
    arr = numpy.ones((5, 5))
    cache.put("a", arr, cost=1)
    assert cache.get("a") is arr
    assert cache.misses == 2
    assert cache.hits == 1
    assert cache.get("a") is arr
    assert cache.misses == 2
    assert cache.hits == 2
    cache.discard("a")
    assert cache.get("a") is None
    assert cache.misses == 3
    assert cache.hits == 2


def test_too_large_item():
    AVAILABLE_BYTES = 10  # very small limit
    cache = ObjectCache(AVAILABLE_BYTES)
    arr = numpy.ones((5, 5))
    assert arr.nbytes > AVAILABLE_BYTES
    cache.put("b", arr, cost=1)
    assert cache.get("b") is None
    # Manually specify the size.
    cache.put("b", arr, cost=1, nbytes=arr.nbytes)
    assert cache.get("b") is None


def test_eviction():
    AVAILABLE_BYTES = 300
    cache = ObjectCache(AVAILABLE_BYTES)
    arr1 = numpy.ones((5, 5))  # 200 bytes
    arr2 = 2 * numpy.ones((5, 5))
    cache.put("arr1", arr1, cost=1)
    assert "arr1" in cache
    # Costly one evicts the previous one.
    cache.put("arr2", arr2, cost=5)
    assert "arr1" not in cache
    assert "arr2" in cache
    # Cheap one does not evict the previous one.
    cache.put("arr1", arr1, cost=1)
    assert "arr1" not in cache
    assert "arr2" in cache


@pytest.mark.asyncio
async def test_object_cache_hit_and_miss(tmpdir):
    with open(Path(tmpdir, "data.csv"), "w") as file:
        file.write(
            """
a,b,c
1,2,3
"""
        )
    adapter = in_memory(readable_storage=[tmpdir])
    with Context.from_app(build_app(adapter)) as context:
        await register(adapter, tmpdir)
        client = from_context(context)
        cache = get_object_cache()
        assert cache.hits == cache.misses == 0
        client["data"].read()
        assert cache.misses == 2  # two dask objects in the cache
        assert cache.hits == 0
        client["data"].read()
        assert cache.misses == 2
        assert cache.hits == 2
        # Simulate eviction.
        cache.clear()
        client["data"].read()
        assert cache.misses == 4
        assert cache.hits == 2
        client["data"].read()
        assert cache.misses == 4
        assert cache.hits == 4


@pytest.mark.asyncio
async def test_object_cache_disabled(tmpdir):
    with open(Path(tmpdir, "data.csv"), "w") as file:
        file.write(
            """
a,b,c
1,2,3
"""
        )
    adapter = in_memory(readable_storage=[tmpdir])
    server_settings = {"object_cache": {"available_bytes": 0}}
    with Context.from_app(
        build_app(adapter, server_settings=server_settings)
    ) as context:
        await register(adapter, tmpdir)
        client = from_context(context)
        cache = get_object_cache()
        assert cache is NO_CACHE
        client["data"]


@pytest.mark.asyncio
async def test_detect_content_changed_or_removed(tmpdir):
    path = Path(tmpdir, "data.csv")
    with open(path, "w") as file:
        file.write(
            """
a,b,c
1,2,3
"""
        )
    adapter = in_memory(readable_storage=[tmpdir])
    with Context.from_app(build_app(adapter)) as context:
        await register(adapter, tmpdir)
        client = from_context(context)
        cache = get_object_cache()
        assert cache.hits == cache.misses == 0
        assert len(client["data"].read()) == 1
        with open(path, "w") as file:
            file.write(
                """
    a,b,c
    1,2,3
    4,5,6
    """
            )
        await register(adapter, tmpdir)
        assert len(client["data"].read()) == 2
        with open(path, "w") as file:
            file.write(
                """
    a,b,c
    1,2,3
    4,5,6
    7,8,9
    """
            )
        await register(adapter, tmpdir)
        assert len(client["data"].read()) == 3
        # Remove file.
        path.unlink()
        await register(adapter, tmpdir)
        assert "data" not in client
        with pytest.raises(KeyError):
            client["data"]


def test_cache_size_absolute(tmpdir):
    config = {
        "trees": [
            {
                "tree": "catalog",
                "path": "/",
                "args": {"uri": tmpdir / "catalog.db", "init_if_not_exists": True},
            }
        ],
        "object_cache": {"available_bytes": 1000},
    }
    with Context.from_app(build_app_from_config(config)):
        cache = get_object_cache()
        assert cache.available_bytes == 1000


def test_cache_size_relative(tmpdir):
    # As a fraction of system memory
    config = {
        "trees": [
            {
                "tree": "catalog",
                "path": "/",
                "args": {"uri": tmpdir / "catalog.db", "init_if_not_exists": True},
            }
        ],
        "object_cache": {"available_bytes": 0.1},
    }
    with Context.from_app(build_app_from_config(config)):
        cache = get_object_cache()
        actual = cache.available_bytes
        expected = psutil.virtual_memory().total * 0.1
        assert abs(actual - expected) / expected < 0.01  # inexact is OK



import io

import numpy
import pytest

from ..adapters import hdf5 as hdf5_adapters
from ..adapters.hdf5 import HDF5Adapter
from ..adapters.mapping import MapAdapter
from ..client import Context, from_context, record_history
from ..server.app import build_app
from ..utils import tree as tree_util


@pytest.fixture
def example_file():
    h5py = pytest.importorskip("h5py")
    file = h5py.File(io.BytesIO(), "w")
    a = file.create_group("a")
    b = a.create_group("b")
    c = b.create_group("c")
    c.create_dataset("d", data=numpy.ones((3, 3)))
    return file


@pytest.fixture
def example_file_with_vlen_str_in_dataset():
    h5py = pytest.importorskip("h5py")
    file = h5py.File(io.BytesIO(), "w")
    a = file.create_group("a")
    b = a.create_group("b")
    c = b.create_group("c")
    # Need to do this to make a vlen str dataset
    dt = h5py.string_dtype(encoding="utf-8")
    dset = c.create_dataset("d", (100,), dtype=dt)
    assert dset.dtype == "object"
    dset[0] = b"test"
    return file


def test_from_file(example_file):
    """Serve a single HDF5 file at top level."""
    h5py = pytest.importorskip("h5py")
    tree = HDF5Adapter(example_file)
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)
        arr = client["a"]["b"]["c"]["d"].read()
        assert isinstance(arr, numpy.ndarray)
        buffer = io.BytesIO()
        client.export(buffer, format="application/x-hdf5")
        file = h5py.File(buffer, "r")
        file["a"]["b"]["c"]["d"]


def test_from_file_with_vlen_str_dataset(example_file_with_vlen_str_in_dataset):
    """Serve a single HDF5 file at top level."""
    h5py = pytest.importorskip("h5py")
    tree = HDF5Adapter(example_file_with_vlen_str_in_dataset)
    with pytest.warns(UserWarning):
        with Context.from_app(build_app(tree)) as context:
            client = from_context(context)
    arr = client["a"]["b"]["c"]["d"].read()
    assert isinstance(arr, numpy.ndarray)
    buffer = io.BytesIO()
    with pytest.warns(UserWarning):
        client.export(buffer, format="application/x-hdf5")
    file = h5py.File(buffer, "r")
    file["a"]["b"]["c"]["d"]


def test_from_group(example_file):
    """Serve a Group within an HDF5 file."""
    h5py = pytest.importorskip("h5py")
    tree = HDF5Adapter(example_file["a"]["b"])
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)
    arr = client["c"]["d"].read()
    assert isinstance(arr, numpy.ndarray)
    buffer = io.BytesIO()
    client.export(buffer, format="application/x-hdf5")
    file = h5py.File(buffer, "r")
    file["c"]["d"]


def test_from_multiple(example_file):
    """Serve two files within a mapping."""
    h5py = pytest.importorskip("h5py")
    tree = MapAdapter(
        {
            "A": HDF5Adapter(example_file),
            "B": HDF5Adapter(example_file),
        }
    )
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)
    arr_A = client["A"]["a"]["b"]["c"]["d"].read()
    assert isinstance(arr_A, numpy.ndarray)
    arr_B = client["B"]["a"]["b"]["c"]["d"].read()
    assert isinstance(arr_B, numpy.ndarray)
    buffer = io.BytesIO()
    client.export(buffer, format="application/x-hdf5")
    file = h5py.File(buffer, "r")
    file["A"]["a"]["b"]["c"]["d"]
    file["B"]["a"]["b"]["c"]["d"]


def test_inlined_contents(example_file):
    """Test that the recursive structure and metadata are inlined into one request."""
    tree = HDF5Adapter(example_file)
    assert hdf5_adapters.INLINED_DEPTH > 1
    original = hdf5_adapters.INLINED_DEPTH
    try:
        with Context.from_app(build_app(tree)) as context:
            with record_history() as hN:
                client = from_context(context)
                tree_util(client)
            hdf5_adapters.INLINED_DEPTH = 1
            with record_history() as h1:
                client = from_context(context)
                tree_util(client)
            hdf5_adapters.INLINED_DEPTH = 0
            with record_history() as h0:
                client = from_context(context)
                tree_util(client)
            assert len(h0.requests) > len(h1.requests) > len(hN.requests)
    finally:
        hdf5_adapters.INLINED_DEPTH = original



"""Add PendingSession.

Revision ID: 4a9dfaba4a98
Revises: 56809bcbfcb0
Create Date: 2022-10-22 19:11:37.926595

"""
from alembic import op
from sqlalchemy import Column, DateTime, ForeignKey, Integer, LargeBinary, Unicode

# revision identifiers, used by Alembic.
revision = "4a9dfaba4a98"
down_revision = "56809bcbfcb0"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "pending_sessions",
        Column(
            "hashed_device_code",
            LargeBinary(32),
            primary_key=True,
            index=True,
            nullable=False,
        ),
        Column("user_code", Unicode(8), index=True, nullable=False),
        Column("expiration_time", DateTime(timezone=False), nullable=False),
        Column("session_id", Integer, ForeignKey("sessions.id"), nullable=True),
    )


def downgrade():
    op.drop_table("pending_sessions")



import dask.array

from ..server.object_cache import get_object_cache
from ..structures.array import ArrayStructure
from ..structures.core import StructureFamily


class ArrayAdapter:
    """
    Wrap an array-like object in an interface that Tiled can serve.

    Examples
    --------

    Wrap any array-like.

    >>> ArrayAdapter.from_array(numpy.random.random((100, 100)))

    >>> ArrayAdapter.from_array(dask.array.from_array(numpy.random.random((100, 100)), chunks=(100, 50)))

    """

    structure_family = StructureFamily.array

    def __init__(
        self,
        array,
        structure,
        *,
        metadata=None,
        specs=None,
        access_policy=None,
    ):
        self._array = array
        self._structure = structure
        self._metadata = metadata or {}
        self.specs = specs or []

    @classmethod
    def from_array(
        cls,
        array,
        *,
        shape=None,
        chunks=None,
        dims=None,
        metadata=None,
        specs=None,
        access_policy=None,
    ):
        structure = ArrayStructure.from_array(
            array, shape=shape, chunks=chunks, dims=dims
        )
        return cls(
            array,
            structure=structure,
            metadata=metadata,
            specs=specs,
            access_policy=access_policy,
        )

    def __repr__(self):
        return f"{type(self).__name__}({self._array!r})"

    @property
    def dims(self):
        return self._structure.dims

    def metadata(self):
        return self._metadata

    def structure(self):
        return self._structure

    def read(self, slice=None):
        array = self._array
        if slice is not None:
            array = array[slice]
        # Special case for dask to cache computed result in object cache.
        if isinstance(self._array, dask.array.Array):
            # Note: If the cache is set to NO_CACHE, this is a null context.
            with get_object_cache().dask_context:
                return array.compute()
        return array

    def read_block(self, block, slice=None):
        # Slice the whole array to get this block.
        slice_, _ = slice_and_shape_from_block_and_chunks(block, self._structure.chunks)
        array = self._array[slice_]
        # Slice within the block.
        if slice is not None:
            array = array[slice]
        # Special case for dask to cache computed result in object cache.
        if isinstance(array, dask.array.Array):
            # Note: If the cache is set to NO_CACHE, this is a null context.
            with get_object_cache().dask_context:
                return array.compute()
        return array


def slice_and_shape_from_block_and_chunks(block, chunks):
    """
    Given dask-like chunks and block id, return slice and shape of the block.
    """
    slice_ = []
    shape = []
    for b, c in zip(block, chunks):
        start = sum(c[:b])
        dim = c[b]
        slice_.append(slice(start, start + dim))
        shape.append(dim)
    return tuple(slice_), tuple(shape)



import io
import sys

import awkward

from ..structures.core import StructureFamily
from ..utils import modules_available

if sys.version_info < (3, 9):
    # Python 3.8 has a bug in zipfile that is not easily matched. Import
    # zipfile from a copy of the Python 3.9.17 version, vendored in the tiled
    # source.
    from . import _zipfile_py39 as zipfile
else:
    import zipfile

from ..media_type_registration import deserialization_registry, serialization_registry
from ..utils import APACHE_ARROW_FILE_MIME_TYPE


@serialization_registry.register(StructureFamily.awkward, "application/zip")
def to_zipped_buffers(components, metadata):
    (form, length, container) = components
    file = io.BytesIO()
    # Pack multiple buffers into a zipfile, uncompressed. This enables
    # multiple buffers in a single response, with random access. The
    # entire payload *may* be compressed using Tiled's normal compression
    # mechanisms.
    with zipfile.ZipFile(file, "w", compresslevel=zipfile.ZIP_STORED) as zip:
        for form_key, buffer in container.items():
            zip.writestr(form_key, buffer)
    return file.getbuffer()


@deserialization_registry.register(StructureFamily.awkward, "application/zip")
def from_zipped_buffers(buffer, form, length):
    file = io.BytesIO(buffer)
    with zipfile.ZipFile(file, "r") as zip:
        form_keys = zip.namelist()
        container = {}
        for form_key in form_keys:
            container[form_key] = zip.read(form_key)
    return container


@serialization_registry.register(StructureFamily.awkward, "application/json")
def to_json(components, metadata):
    (form, length, container) = components
    file = io.StringIO()
    array = awkward.from_buffers(form, length, container)
    awkward.to_json(array, file)
    return file.getvalue()


if modules_available("pyarrow"):

    @serialization_registry.register(
        StructureFamily.awkward, APACHE_ARROW_FILE_MIME_TYPE
    )
    def to_arrow(components, metadata):
        import pyarrow

        (form, length, container) = components
        array = awkward.from_buffers(form, length, container)
        table = awkward.to_arrow_table(array)
        sink = pyarrow.BufferOutputStream()
        with pyarrow.ipc.new_file(sink, table.schema) as writer:
            writer.write_table(table)
        return memoryview(sink.getvalue())

    # There seems to be no official Parquet MIME type.
    # https://issues.apache.org/jira/browse/PARQUET-1889
    @serialization_registry.register(StructureFamily.awkward, "application/x-parquet")
    def to_parquet(components, metadata):
        import pyarrow.parquet

        (form, length, container) = components
        array = awkward.from_buffers(form, length, container)
        table = awkward.to_arrow_table(array)
        sink = pyarrow.BufferOutputStream()
        with pyarrow.parquet.ParquetWriter(sink, table.schema) as writer:
            writer.write_table(table)
        return memoryview(sink.getvalue())



from pathlib import Path

import h5py
import numpy

from ..server.etag import tokenize


def test_deterministic_h5py_dataset(tmpdir):
    file = h5py.File(Path(tmpdir, "test.h5"), "w")
    group = file.create_group("stuff")
    dataset = group.create_dataset("data", data=numpy.ones((3, 3)))
    assert tokenize(dataset) == tokenize(dataset)



import io
import json

import awkward
import numpy
import pyarrow.feather
import pyarrow.parquet
import pytest

from ..catalog import in_memory
from ..client import Context, from_context, record_history
from ..server.app import build_app
from ..utils import APACHE_ARROW_FILE_MIME_TYPE


@pytest.fixture
def catalog(tmpdir):
    catalog = in_memory(writable_storage=tmpdir)
    yield catalog


@pytest.fixture
def app(catalog):
    app = build_app(catalog)
    yield app


@pytest.fixture
def context(app):
    with Context.from_app(app) as context:
        yield context


@pytest.fixture
def client(context):
    client = from_context(context)
    yield client


def test_slicing(client):
    # Write data into catalog. It will be stored as directory of buffers
    # named like 'node0-offsets' and 'node2-data'.
    array = awkward.Array(
        [
            [{"x": 1.1, "y": [1]}, {"x": 2.2, "y": [1, 2]}],
            [],
            [{"x": 3.3, "y": [1, 2, 3]}],
        ]
    )
    returned = client.write_awkward(array, key="test")
    # Test with client returned, and with client from lookup.
    for aac in [returned, client["test"]]:
        # Read the data back out from the AwkwardArrrayClient, progressively sliced.
        assert awkward.almost_equal(aac.read(), array)
        assert awkward.almost_equal(aac[:], array)
        assert awkward.almost_equal(aac[0], array[0])
        assert awkward.almost_equal(aac[0, "y"], array[0, "y"])
        assert awkward.almost_equal(aac[0, "y", :1], array[0, "y", :1])

        # When sliced, the serer sends less data.
        with record_history() as h:
            aac[:]
        assert len(h.responses) == 1  # sanity check
        full_response_size = len(h.responses[0].content)
        with record_history() as h:
            aac[0, "y"]
        assert len(h.responses) == 1  # sanity check
        sliced_response_size = len(h.responses[0].content)
        assert sliced_response_size < full_response_size


def test_export_json(client):
    # Write data into catalog. It will be stored as directory of buffers
    # named like 'node0-offsets' and 'node2-data'.
    array = awkward.Array(
        [
            [{"x": 1.1, "y": [1]}, {"x": 2.2, "y": [1, 2]}],
            [],
            [{"x": 3.3, "y": [1, 2, 3]}],
        ]
    )
    aac = client.write_awkward(array, key="test")

    file = io.BytesIO()
    aac.export(file, format="application/json")
    actual = bytes(file.getbuffer()).decode()
    assert actual == awkward.to_json(array)


def test_export_arrow(tmpdir, client):
    # Write data into catalog. It will be stored as directory of buffers
    # named like 'node0-offsets' and 'node2-data'.
    array = awkward.Array(
        [
            [{"x": 1.1, "y": [1]}, {"x": 2.2, "y": [1, 2]}],
            [],
            [{"x": 3.3, "y": [1, 2, 3]}],
        ]
    )
    aac = client.write_awkward(array, key="test")

    filepath = tmpdir / "actual.arrow"
    aac.export(str(filepath), format=APACHE_ARROW_FILE_MIME_TYPE)
    actual = pyarrow.feather.read_table(filepath)
    expected = awkward.to_arrow_table(array)
    assert actual == expected


def test_export_parquet(tmpdir, client):
    # Write data into catalog. It will be stored as directory of buffers
    # named like 'node0-offsets' and 'node2-data'.
    array = awkward.Array(
        [
            [{"x": 1.1, "y": [1]}, {"x": 2.2, "y": [1, 2]}],
            [],
            [{"x": 3.3, "y": [1, 2, 3]}],
        ]
    )
    aac = client.write_awkward(array, key="test")

    filepath = tmpdir / "actual.parquet"
    aac.export(str(filepath), format="application/x-parquet")
    actual = pyarrow.parquet.read_table(filepath)
    expected = awkward.to_arrow_table(array)
    assert actual == expected


def test_large_number_of_form_keys(client):
    "Request should succeed even when too many form_keys to fit in a URL."
    # https://github.com/bluesky/tiled/pull/577

    # The HTTP spec itself has no limit, but tools impose a pragmatic one.
    # Nginx defaults to 8k bytes. Chrome is around 2000.
    # This is a representative value.
    URL_LENGTH_LIMIT = 2000
    NUM_KEYS = 7000  # meant to achieve a large body of form_keys
    array = awkward.Array([{f"key{i:05}": i for i in range(NUM_KEYS)}])
    aac = client.write_awkward(array, key="test")
    with record_history() as h:
        aac[0]
    (request,) = h.requests
    assert request.method == "POST"
    form_keys = json.loads(request.read())
    form_key_length = len(str(form_keys))
    assert form_key_length > URL_LENGTH_LIMIT
    url_length = len(str(request.url))
    assert url_length <= URL_LENGTH_LIMIT


def test_more_slicing_1(client):
    array = awkward.Array(
        [
            {
                "stuff": 123,
                "file": [
                    {"filename": 321, "other": 3.14},
                ],
            },
        ]
    )
    returned = client.write_awkward(array, key="test")
    # Test with client returned, and with client from lookup.
    for aac in [returned, client["test"]]:
        # Read the data back out from the AwkwardArrrayClient, progressively sliced.
        assert awkward.almost_equal(aac.read(), array)
        assert awkward.almost_equal(aac[:], array)
        assert awkward.almost_equal(aac["file", "filename"], array["file", "filename"])


def test_more_slicing_2(client):
    array = awkward.from_numpy(
        numpy.arange(2 * 3 * 5).reshape(2, 3, 5), regulararray=True
    )
    returned = client.write_awkward(array, key="test")
    # Test with client returned, and with client from lookup.
    for aac in [returned, client["test"]]:
        # Read the data back out from the AwkwardArrrayClient, progressively sliced.
        assert awkward.almost_equal(aac.read(), array)
        assert awkward.almost_equal(aac[:], array)
        assert awkward.almost_equal(aac[1:], array[1:])
        assert awkward.almost_equal(aac[1:, 1:], array[1:, 1:])


def test_more_slicing_3(client):
    array = awkward.Array(
        [
            {"good": 123, "bad": 123},
            {"good": 321, "bad": [1, 2, 3]},
        ]
    )
    returned = client.write_awkward(array, key="test")
    # Test with client returned, and with client from lookup.
    for aac in [returned, client["test"]]:
        # Read the data back out from the AwkwardArrrayClient, progressively sliced.
        assert awkward.almost_equal(aac.read(), array)
        assert awkward.almost_equal(aac[:], array[:])
        assert awkward.almost_equal(aac["good"], array["good"])



import io

from ..media_type_registration import serialization_registry
from ..utils import ensure_awaitable, modules_available
from .container import walk
from .table import (
    APACHE_ARROW_FILE_MIME_TYPE,
    XLSX_MIME_TYPE,
    serialize_arrow,
    serialize_csv,
    serialize_excel,
    serialize_html,
    serialize_parquet,
)


async def as_dataset(node):
    import xarray

    data_vars = {}
    coords = {}
    if hasattr(node, "items_range"):
        items = await node.items_range(0, None)
    else:
        items = node.items()
    for key, array_adapter in items:
        spec_names = set(spec.name for spec in array_adapter.specs)
        arr = await ensure_awaitable(array_adapter.read)
        if "xarray_data_var" in spec_names:
            data_vars[key] = (array_adapter.structure().dims, arr)
        elif "xarray_coord" in spec_names:
            coords[key] = (array_adapter.structure().dims, arr)
        else:
            raise ValueError(
                "Child nodes of xarray_dataset should include spec "
                "'xarray_coord' or 'xarray_data_var'."
            )
    return xarray.Dataset(
        data_vars=data_vars, coords=coords, attrs=node.metadata()["attrs"]
    )


class _BytesIOThatIgnoresClose(io.BytesIO):
    def close(self):
        # When the netcdf writer tells us to close(), ignore it.
        pass


if modules_available("scipy"):
    # Both application/netcdf and application/x-netcdf are used.
    # https://en.wikipedia.org/wiki/NetCDF
    @serialization_registry.register(
        "xarray_dataset", ["application/netcdf", "application/x-netcdf"]
    )
    async def serialize_netcdf(node, metadata, filter_for_access):
        file = _BytesIOThatIgnoresClose()
        # Per the xarray.Dataset.to_netcdf documentation,
        # file-like objects are only supported by the scipy engine.
        (await as_dataset(node)).to_netcdf(file, engine="scipy")
        return file.getbuffer()


# Support DataFrame formats by first converting to DataFrame.
# This doesn't make much sense for N-dimensional variables, but for
# 1-dimensional variables it is useful.


@serialization_registry.register("xarray_dataset", APACHE_ARROW_FILE_MIME_TYPE)
async def serialize_dataset_arrow(node, metadata, filter_for_access):
    return serialize_arrow((await as_dataset(node)).to_dataframe(), metadata)


@serialization_registry.register("xarray_dataset", "application/x-parquet")
async def serialize_dataset_parquet(node, metadata, filter_for_access):
    return serialize_parquet((await as_dataset(node)).to_dataframe(), metadata)


@serialization_registry.register(
    "xarray_dataset", ["text/csv", "text/comma-separated-values", "text/plain"]
)
async def serialize_dataset_csv(node, metadata, filter_for_access):
    return serialize_csv((await as_dataset(node)).to_dataframe(), metadata)


@serialization_registry.register("xarray_dataset", "text/html")
async def serialize_dataset_html(node, metadata, filter_for_access):
    return serialize_html((await as_dataset(node)).to_dataframe(), metadata)


@serialization_registry.register("xarray_dataset", XLSX_MIME_TYPE)
async def serialize_dataset_excel(node, metadata, filter_for_access):
    return serialize_excel((await as_dataset(node)).to_dataframe(), metadata)


if modules_available("orjson"):
    import orjson

    @serialization_registry.register("xarray_dataset", "application/json")
    async def serialize_json(node, metadata, filter_for_access):
        df = (await as_dataset(node)).to_dataframe()
        return orjson.dumps(
            {column: df[column].tolist() for column in df},
        )


if modules_available("h5py"):

    @serialization_registry.register("xarray_dataset", "application/x-hdf5")
    async def serialize_hdf5(node, metadata, filter_for_access):
        """
        Like for node, but encode everything under 'attrs' in attrs.
        """
        import h5py

        buffer = io.BytesIO()
        root_node = node
        with h5py.File(buffer, mode="w") as file:
            for k, v in metadata["attrs"].items():
                file.attrs.create(k, v)
            async for key_path, array_adapter in walk(node, filter_for_access):
                group = file
                node = root_node
                for key in key_path[:-1]:
                    node = node[key]
                    if key in group:
                        group = group[key]
                    else:
                        group = group.create_group(key)
                        group.attrs.update(node.metadata()["attrs"])
                data = array_adapter.read()
                dataset = group.create_dataset(key_path[-1], data=data)
                for k, v in array_adapter.metadata()["attrs"].items():
                    dataset.attrs.create(k, v)
        return buffer.getbuffer()



from ..utils import tree  # noqa: F401
from .constructors import from_context, from_profile, from_uri  # noqa: F401
from .container import ASCENDING, DESCENDING  # noqa: F401
from .context import Context  # noqa: F401
from .logger import hide_logs, record_history, show_logs  # noqa: F401



import hashlib
import re
from pathlib import Path
from urllib.parse import urlparse, urlunparse

import canonicaljson

SCHEME_PATTERN = re.compile(r"^[a-z0-9+]+:\/\/.*$")


def ensure_uri(uri_or_path):
    "Accept a URI or file path (Windows- or POSIX-style) and return a URI."
    if not SCHEME_PATTERN.match(str(uri_or_path)):
        # Interpret this as a filepath.
        path = uri_or_path
        uri_str = urlunparse(
            ("file", "localhost", str(Path(path).absolute()), "", "", None)
        )
    else:
        # Interpret this as a URI.
        uri_str = uri_or_path
        parsed = urlparse(uri_str)
        if parsed.netloc == "":
            mutable = list(parsed)
            mutable[1] = "localhost"
            uri_str = urlunparse(mutable)
    return str(uri_str)


def compute_structure_id(structure):
    "Compute HEX digest of MD5 hash of RFC 8785 canonical form of JSON."
    canonical_structure = canonicaljson.encode_canonical_json(structure)

    return hashlib.md5(canonical_structure).hexdigest()



import dask.dataframe

from ..server.object_cache import NO_CACHE, get_object_cache
from ..utils import path_from_uri
from .dataframe import DataFrameAdapter


def read_csv(
    data_uri,
    structure=None,
    metadata=None,
    specs=None,
    access_policy=None,
    **kwargs,
):
    """
    Read a CSV.

    Internally, this uses dask.dataframe.read_csv.
    It forward all parameters to that function. See
    https://docs.dask.org/en/latest/dataframe-api.html#dask.dataframe.read_csv

    Examples
    --------

    >>> read_csv("myfiles.*.csv")
    >>> read_csv("s3://bucket/myfiles.*.csv")
    """
    filepath = path_from_uri(data_uri)
    ddf = dask.dataframe.read_csv(filepath, **kwargs)
    # If an instance has previously been created using the same parameters,
    # then we are here because the caller wants a *fresh* view on this data.
    # Therefore, we should clear any cached data.
    cache = get_object_cache()
    if cache is not NO_CACHE:
        cache.discard_dask(ddf.__dask_keys__())
    # TODO Pass structure through rather than just re-creating it
    # in from_dask_dataframe.
    return DataFrameAdapter.from_dask_dataframe(
        ddf, metadata=metadata, specs=specs, access_policy=access_policy
    )


read_csv.__doc__ = (
    """
This wraps dask.dataframe.read_csv. Original docstring:

"""
    + dask.dataframe.read_csv.__doc__
)



import asyncio
import re
from pathlib import Path
from typing import List, Optional

import typer

from ._serve import serve_catalog

catalog_app = typer.Typer()
# Support both `tiled serve catalog` and `tiled catalog serve` as synonyms
# because I cannot decide which is right.
catalog_app.command("serve")(serve_catalog)
DEFAULT_SQLITE_CATALOG_FILENAME = "catalog.db"


@catalog_app.command("init")
def init(
    database: str = typer.Argument(
        Path.cwd() / DEFAULT_SQLITE_CATALOG_FILENAME, help="A filepath or database URI"
    ),
    if_not_exists: bool = typer.Option(
        False,
        help=(
            "By default, it is an error if a database is already initialized."
            "Set this flag to be permissive and return without an error."
        ),
    ),
):
    """
    Initialize a database as a Tiled Catalog.

    Examples:

    # Using a simple local file as an embedded "database" (SQLite)
    tiled init catalog.db
    tiled init path/to/catalog.db
    tiled init sqlite+aiosqlite:////path/to/catalog.db

    # Using a client/serve database engine (PostgreSQL)
    tiled init postgresql+asyncpg://uesrname:password@localhost/database_name:5432
    """
    from sqlalchemy.ext.asyncio import create_async_engine

    from ..alembic_utils import UninitializedDatabase, check_database, stamp_head
    from ..catalog.alembic_constants import ALEMBIC_DIR, ALEMBIC_INI_TEMPLATE_PATH
    from ..catalog.core import ALL_REVISIONS, REQUIRED_REVISION, initialize_database
    from ..catalog.utils import SCHEME_PATTERN

    if not SCHEME_PATTERN.match(database):
        # Interpret URI as filepath.
        database = f"sqlite+aiosqlite:///{database}"

    async def do_setup():
        engine = create_async_engine(database)
        redacted_url = engine.url._replace(password="[redacted]")
        try:
            await check_database(engine, REQUIRED_REVISION, ALL_REVISIONS)
        except UninitializedDatabase:
            # Create tables and stamp (alembic) revision.
            typer.echo(
                f"Database {redacted_url} is new. Creating tables.",
                err=True,
            )
            await initialize_database(engine)
            typer.echo("Database initialized.", err=True)
        else:
            if not if_not_exists:
                typer.echo(
                    f"Database at {redacted_url} is already initialized.", err=True
                )
                raise typer.Abort()
        finally:
            await engine.dispose()

    asyncio.run(do_setup())
    stamp_head(ALEMBIC_INI_TEMPLATE_PATH, ALEMBIC_DIR, database)


@catalog_app.command("register")
def register(
    database: str = typer.Argument(
        ..., help="A filepath or database URI for the catalog"
    ),
    filepath: str = typer.Argument(..., help="A file or directory to register"),
    verbose: bool = typer.Option(
        False,
        "--verbose",
        "-v",
        help=("Log details of directory traversal and file registration."),
    ),
    watch: bool = typer.Option(
        False,
        "--watch",
        "-w",
        help="Update catalog when files are added, removed, or changed.",
    ),
    prefix: str = typer.Option(
        "/", help="Location within the catalog's namespace to register these files"
    ),
    keep_ext: bool = typer.Option(
        False,
        "--keep-ext",
        help=(
            "Serve a file like 'measurements.csv' as its full filepath with extension, "
            "instead of the default which would serve it as 'measurements'. "
            "This is discouraged because it leaks details about the storage "
            "format to the client, such that changing the storage in the future "
            "may break user (client-side) code."
        ),
    ),
    ext: List[str] = typer.Option(
        None,
        "--ext",
        help=(
            "Support custom file extension, mapping it to a known mimetype. "
            "Spell like '.tif=image/tiff'. Include the leading '.' in the file "
            "extension."
        ),
    ),
    mimetype_detection_hook: str = typer.Option(
        None,
        "--mimetype-hook",
        help=(
            "ADVANCED: Custom mimetype detection Python function. "
            "Expected interface: detect_mimetype(filepath, mimetype) -> mimetype "
            "Specify here as 'package.module:function'"
        ),
    ),
    adapters: List[str] = typer.Option(
        None,
        "--adapter",
        help=(
            "ADVANCED: Custom Tiled Adapter for reading a given format"
            "Specify here as 'mimetype=package.module:function'"
        ),
    ),
    walkers: List[str] = typer.Option(
        None,
        "--walker",
        help=(
            "ADVANCED: Custom Tiled Walker for traversing directories and "
            "grouping files. This is used in conjunction with Adapters that operate "
            "on groups of files. "
            "Specify here as 'package.module:function'"
        ),
    ),
):
    from ..catalog.utils import SCHEME_PATTERN

    if not SCHEME_PATTERN.match(database):
        # Interpret URI as filepath.
        database = f"sqlite+aiosqlite:///{database}"
    if keep_ext:
        from ..catalog.register import identity

        key_from_filename = identity
    else:
        key_from_filename = None
    mimetypes_by_file_ext = {}
    EXT_PATTERN = re.compile(r"(.*) *= *(.*)")
    for item in ext or []:
        match = EXT_PATTERN.match(item)
        if match is None:
            raise ValueError(
                f"Failed parsing --ext option {item}, expected format '.ext=mimetype'"
            )
        ext, mimetype = match.groups()
        mimetypes_by_file_ext[ext] = mimetype
    adapters_by_mimetype = {}
    ADAPTER_PATTERN = re.compile(r"(.*) *= *(.*)")
    for item in adapters or []:
        match = ADAPTER_PATTERN.match(item)
        if match is None:
            raise ValueError(
                f"Failed parsing --adapter option {item}, expected format 'mimetype=package.module:obj'"
            )
        mimetype, obj_ref = match.groups()
        adapters_by_mimetype[mimetype] = obj_ref

    from ..catalog import from_uri

    catalog_adapter = from_uri(database)

    from logging import StreamHandler

    from ..catalog.register import logger as register_logger
    from ..catalog.register import register
    from ..catalog.register import watch as watch_

    if verbose:
        register_logger.addHandler(StreamHandler())
        register_logger.setLevel("INFO")
    if watch:
        asyncio.run(
            watch_(
                catalog_adapter,
                filepath,
                prefix=prefix,
                mimetype_detection_hook=mimetype_detection_hook,
                mimetypes_by_file_ext=mimetypes_by_file_ext,
                adapters_by_mimetype=adapters_by_mimetype,
                walkers=walkers,
                key_from_filename=key_from_filename,
            )
        )
    else:
        asyncio.run(
            register(
                catalog_adapter,
                filepath,
                prefix=prefix,
                mimetype_detection_hook=mimetype_detection_hook,
                mimetypes_by_file_ext=mimetypes_by_file_ext,
                adapters_by_mimetype=adapters_by_mimetype,
                walkers=walkers,
                key_from_filename=key_from_filename,
            )
        )


@catalog_app.command("upgrade-database")
def upgrade_database(
    database_uri: str,
    revision: Optional[str] = typer.Argument(
        None,
        help="The ID of a revision to upgrade to. By default, upgrade to the latest one.",
    ),
):
    """
    Upgrade the catalog database schema to the latest version.
    """
    import asyncio

    from sqlalchemy.ext.asyncio import create_async_engine

    from ..alembic_utils import get_current_revision, upgrade
    from ..catalog.alembic_constants import ALEMBIC_DIR, ALEMBIC_INI_TEMPLATE_PATH
    from ..catalog.core import ALL_REVISIONS

    async def do_setup():
        engine = create_async_engine(database_uri)
        redacted_url = engine.url._replace(password="[redacted]")
        current_revision = await get_current_revision(engine, ALL_REVISIONS)
        await engine.dispose()
        if current_revision is None:
            # Create tables and stamp (alembic) revision.
            typer.echo(
                f"Database {redacted_url} has not been initialized. Use `tiled catalog init`.",
                err=True,
            )
            raise typer.Abort()

    asyncio.run(do_setup())
    upgrade(ALEMBIC_INI_TEMPLATE_PATH, ALEMBIC_DIR, database_uri, revision or "head")


@catalog_app.command("downgrade-database")
def downgrade_database(
    database_uri: str,
    revision: str = typer.Argument(..., help="The ID of a revision to downgrade to."),
):
    """
    Upgrade the catalog database schema to the latest version.
    """
    import asyncio

    from sqlalchemy.ext.asyncio import create_async_engine

    from ..alembic_utils import downgrade, get_current_revision
    from ..catalog.alembic_constants import ALEMBIC_DIR, ALEMBIC_INI_TEMPLATE_PATH
    from ..catalog.core import ALL_REVISIONS

    async def do_setup():
        engine = create_async_engine(database_uri)
        redacted_url = engine.url._replace(password="[redacted]")
        current_revision = await get_current_revision(engine, ALL_REVISIONS)
        if current_revision is None:
            # Create tables and stamp (alembic) revision.
            typer.echo(
                f"Database {redacted_url} has not been initialized. Use `tiled catalog init`.",
                err=True,
            )
            raise typer.Abort()

    asyncio.run(do_setup())
    downgrade(ALEMBIC_INI_TEMPLATE_PATH, ALEMBIC_DIR, database_uri, revision)



from pathlib import Path

import numpy
import pytest
import tifffile as tf

from ..adapters.mapping import MapAdapter
from ..adapters.tiff import TiffAdapter, TiffSequenceAdapter
from ..catalog import in_memory
from ..catalog.register import TIFF_SEQUENCE_EMPTY_NAME_ROOT, register
from ..catalog.utils import ensure_uri
from ..client import Context, from_context
from ..server.app import build_app

COLOR_SHAPE = (11, 17, 3)


@pytest.fixture(scope="module")
def client(tmpdir_module):
    sequence_directory = Path(tmpdir_module, "sequence")
    sequence_directory.mkdir()
    filepaths = []
    for i in range(3):
        data = numpy.random.random((5, 7))
        filepath = sequence_directory / f"temp{i:05}.tif"
        tf.imwrite(filepath, data)
        filepaths.append(filepath)
    color_data = numpy.random.randint(0, 255, COLOR_SHAPE, dtype="uint8")
    path = Path(tmpdir_module, "color.tif")
    tf.imwrite(path, color_data)

    tree = MapAdapter(
        {
            "color": TiffAdapter(ensure_uri(path)),
            "sequence": TiffSequenceAdapter.from_uris(
                [ensure_uri(filepath) for filepath in filepaths]
            ),
        }
    )
    app = build_app(tree)
    with Context.from_app(app) as context:
        client = from_context(context)
        yield client


@pytest.mark.parametrize(
    "slice_input, correct_shape",
    [
        (None, (3, 5, 7)),
        (0, (5, 7)),
        (slice(0, 3, 2), (2, 5, 7)),
        ((1, slice(0, 3), slice(0, 3)), (3, 3)),
        ((slice(0, 3), slice(0, 3), slice(0, 3)), (3, 3, 3)),
    ],
)
def test_tiff_sequence(client, slice_input, correct_shape):
    arr = client["sequence"].read(slice=slice_input)
    assert arr.shape == correct_shape


@pytest.mark.parametrize("block_input, correct_shape", [((0, 0, 0), (1, 5, 7))])
def test_tiff_sequence_block(client, block_input, correct_shape):
    arr = client["sequence"].read_block(block_input)
    assert arr.shape == correct_shape


@pytest.mark.asyncio
async def test_tiff_sequence_order(tmpdir):
    """
    directory/
      00001.tif
      00002.tif
      ...
      00010.tif
    """
    data = numpy.ones((4, 5))
    num_files = 10
    for i in range(num_files):
        tf.imwrite(Path(tmpdir / f"image{i:05}.tif"), data * i)

    adapter = in_memory(readable_storage=[tmpdir])
    with Context.from_app(build_app(adapter)) as context:
        await register(adapter, tmpdir)
        client = from_context(context)
        for i in range(num_files):
            numpy.testing.assert_equal(client["image"][i], data * i)


@pytest.mark.asyncio
async def test_tiff_sequence_with_directory_walker(tmpdir):
    """
    directory/
      00001.tif
      00002.tif
      ...
      00010.tif
      single_image.tif
      image00001.tif
      image00002.tif
      ...
      image00010.tif
      other_image00001.tif
      other_image00002.tif
      ...
      other_image00010.tif
      other_image2_00001.tif
      other_image2_00002.tif
      ...
      other_image2_00010.tif
      other_file1.csv
      other_file2.csv
      stuff.csv
    """
    data = numpy.random.random((3, 5))
    for i in range(10):
        tf.imwrite(Path(tmpdir / f"image{i:05}.tif"), data)
        tf.imwrite(Path(tmpdir / f"other_image{i:05}.tif"), data)
        tf.imwrite(Path(tmpdir / f"{i:05}.tif"), data)
        tf.imwrite(Path(tmpdir / f"other_image2_{i:05}.tif"), data)
    tf.imwrite(Path(tmpdir / "single_image.tif"), data)
    for target in ["stuff.csv", "other_file1.csv", "other_file2.csv"]:
        with open(Path(tmpdir / target), "w") as file:
            file.write(
                """
a,b,c
1,2,3
"""
            )
    adapter = in_memory(readable_storage=[tmpdir])
    with Context.from_app(build_app(adapter)) as context:
        await register(adapter, tmpdir)
        client = from_context(context)
        # Single image is its own node.
        assert client["single_image"].shape == (3, 5)
        # Each sequence is grouped into a node.
        assert client[TIFF_SEQUENCE_EMPTY_NAME_ROOT].shape == (10, 3, 5)
        assert client["image"].shape == (10, 3, 5)
        assert client["other_image"].shape == (10, 3, 5)
        assert client["other_image2_"].shape == (10, 3, 5)
        # The sequence grouping digit-only files appears with a uuid
        named_keys = [
            "single_image",
            "image",
            "other_image",
            "other_image2_",
            "other_file1",
            "other_file2",
            "stuff",
        ]
        no_name_keys = [key for key in client.keys() if key not in named_keys]
        # There is only a single one of this type
        assert len(no_name_keys) == 1
        assert client[no_name_keys[0]].shape == (10, 3, 5)
        # Other files are single nodes.
        assert client["stuff"].columns == ["a", "b", "c"]
        assert client["other_file1"].columns == ["a", "b", "c"]
        assert client["other_file2"].columns == ["a", "b", "c"]


def test_rgb(client):
    "Test an RGB TIFF."
    arr = client["color"].read()
    assert arr.shape == COLOR_SHAPE


def test_tiff_sequence_cache(client):
    from numpy.testing import assert_raises

    # The two requests go through the same method in the server (read_block) to
    # call the same object
    indexed_array = client["sequence"][0]
    read_array = client["sequence"].read(0)

    # Using a different index to confirm that the previous cache doesn't affect the new array
    other_read_array = client["sequence"].read(1)

    numpy.testing.assert_equal(indexed_array, read_array)
    assert_raises(
        AssertionError, numpy.testing.assert_equal, read_array, other_read_array
    )



import dask.base
import dask.dataframe
import pandas

from ..server.object_cache import get_object_cache
from ..structures.core import Spec, StructureFamily
from ..structures.table import TableStructure
from .array import ArrayAdapter


class TableAdapter:
    """
    Wrap a dataframe-like object in an interface that Tiled can serve.

    Examples
    --------

    >>> df = pandas.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    >>> DataFrameAdapter.from_pandas(df, npartitions=1)

    """

    structure_family = StructureFamily.table

    @classmethod
    def from_pandas(
        cls,
        *args,
        metadata=None,
        specs=None,
        access_policy=None,
        npartitions=1,
        **kwargs,
    ):
        ddf = dask.dataframe.from_pandas(*args, npartitions=npartitions, **kwargs)
        if specs is None:
            specs = [Spec("dataframe")]
        return cls.from_dask_dataframe(
            ddf, metadata=metadata, specs=specs, access_policy=access_policy
        )

    @classmethod
    def from_dask_dataframe(
        cls,
        ddf,
        metadata=None,
        specs=None,
        access_policy=None,
    ):
        structure = TableStructure.from_dask_dataframe(ddf)
        if specs is None:
            specs = [Spec("dataframe")]
        return cls(
            ddf.partitions,
            structure,
            metadata=metadata,
            specs=specs,
            access_policy=access_policy,
        )

    def __init__(
        self,
        partitions,
        structure,
        *,
        metadata=None,
        specs=None,
        access_policy=None,
    ):
        self._metadata = metadata or {}
        self._partitions = list(partitions)
        self._structure = structure
        self.specs = specs or []
        self.access_policy = access_policy

    def __repr__(self):
        return f"{type(self).__name__}({self._structure.columns!r})"

    def __getitem__(self, key):
        # Must compute to determine shape.
        return ArrayAdapter.from_array(self.read([key])[key].values)

    def items(self):
        yield from (
            (key, ArrayAdapter.from_array(self.read([key])[key].values))
            for key in self._structure.columns
        )

    def metadata(self):
        return self._metadata

    def structure(self):
        return self._structure

    def read(self, fields=None):
        if any(p is None for p in self._partitions):
            raise ValueError("Not all partitions have been stored.")
        if isinstance(self._partitions[0], dask.dataframe.DataFrame):
            if fields is not None:
                ddf = dask.dataframe.concat(
                    [p[fields] for p in self._partitions], axis=0
                )
            else:
                ddf = dask.dataframe.concat(self._partitions, axis=0)
            # Note: If the cache is set to NO_CACHE, this is a null context.
            with get_object_cache().dask_context:
                return ddf.compute()
        df = pandas.concat(self._partitions, axis=0)
        if fields is not None:
            df = df[fields]
        return df

    def read_partition(self, partition, fields=None):
        partition = self._partitions[partition]
        if partition is None:
            raise RuntimeError(f"partition {partition} has not be stored yet")
        if fields is not None:
            partition = partition[fields]
        # Special case for dask to cache computed result in object cache.
        if isinstance(partition, dask.dataframe.DataFrame):
            # Note: If the cache is set to NO_CACHE, this is a null context.
            with get_object_cache().dask_context:
                return partition.compute()
        return partition


DataFrameAdapter = TableAdapter



"""
Test the feature wherein the descriptions (metadata, structure_family,
structure, etc.) for a node's children may be included with the
description of a node. This is used in xarray_dataset now and may
be used more widely later.
"""

import numpy
import pytest
import xarray

from ..adapters.array import ArrayAdapter
from ..adapters.mapping import MapAdapter
from ..adapters.xarray import DatasetAdapter
from ..client import Context, from_context, record_history
from ..server.app import build_app
from ..server.core import INLINED_CONTENTS_LIMIT

tree = MapAdapter(
    {
        "dataset": DatasetAdapter.from_dataset(
            xarray.Dataset(
                data_vars={"temperature": ("time", [100, 99, 98])},
                coords={"time": [1, 2, 3]},
            )
        ),
    },
    metadata={"thing": "stuff"},
)


@pytest.fixture(scope="module")
def client():
    app = build_app(tree)
    with Context.from_app(app) as context:
        client = from_context(context)
        yield client


def test_lookup(client):
    "Accessing a child node should trigger one request."
    with record_history() as history:
        client["dataset"]
    assert len(history.requests) == 1


def test_iter(client):
    "Iteration should be free because the contents were in-lined."
    expected = ["temperature", "time"]
    dsc = client["dataset"]
    with record_history() as history:
        assert list(dsc) == expected
    assert not history.requests

    # Implementation detail:
    # Without inlined contents, a request is needed.
    with record_history() as history:
        assert list(dsc.__iter__(_ignore_inlined_contents=True)) == expected
    assert history.requests


def test_keys_slice(client):
    "Iteration should be free because the contents were in-lined."
    expected = ["temperature", "time"]
    dsc = client["dataset"]
    with record_history() as history:
        assert list(dsc.keys()) == expected
    assert not history.requests

    with record_history() as history:
        assert dsc.keys()[:] == expected
    assert not history.requests

    with record_history() as history:
        assert dsc.keys().first() == "temperature"
        assert dsc.keys().last() == "time"
    assert not history.requests

    # Implementation detail:
    # Without inlined contents, a request is needed.
    with record_history() as history:
        list(dsc._keys_slice(0, 1, 1, _ignore_inlined_contents=True))
    assert history.requests


def test_items_slice(client):
    "Iteration should be free because the contents were in-lined."
    dsc = client["dataset"]
    with record_history() as history:
        list(dsc.items())
    assert not history.requests

    with record_history() as history:
        dsc.items()
    assert not history.requests

    with record_history() as history:
        dsc.items().first()
    assert not history.requests

    with record_history() as history:
        dsc.items().last()
    assert not history.requests

    # Implementation detail:
    # Without inlined contents, a request is needed.
    with record_history() as history:
        list(dsc._items_slice(0, 1, 1, _ignore_inlined_contents=True))
    assert history.requests


def test_too_wide_for_inline():
    """
    The server will not inline contents above a certain limit.


    It is fetched in pages on demand, as usual with nodes.
    """

    class Adapter(MapAdapter):
        def inlined_contents_enabled(self):
            return True

    a = numpy.array([1])
    tree = MapAdapter(
        {
            f"item{i:05}": ArrayAdapter.from_array(i * a)
            for i in range(1 + 2 * INLINED_CONTENTS_LIMIT)
        }
    )
    app = build_app(tree)
    with Context.from_app(app) as context:
        client = from_context(context)
        assert client.item["attributes"]["structure"]["contents"] is None
        with record_history() as history:
            assert set(client) == set(tree)
        assert len(history.requests) >= 3



"""
The 'data' cache is available to all Adapters to cache chunks of data.

This is integrated with dask's callback mechanism for simple caching of dask
chunks. It is also available for usage outside of dask.

The cache is a process-global singleton.
"""
import contextlib
import time
from timeit import default_timer

import cachey
from dask.callbacks import Callback

if __debug__:
    import logging

    logger = logging.getLogger(__name__)
    handler = logging.StreamHandler()
    handler.setLevel("DEBUG")
    handler.setFormatter(logging.Formatter("OBJECT CACHE: %(message)s"))
    logger.addHandler(handler)


class _NO_CACHE_SENTINEL:
    def __init__(self):
        self.dask_context = contextlib.nullcontext()

    def __repr__(self):
        return "NO_CACHE"


NO_CACHE = _NO_CACHE_SENTINEL()
_object_cache = NO_CACHE


def set_object_cache(cache):
    """
    Set the process-global icache.
    """
    global _object_cache
    _object_cache = cache


def get_object_cache():
    """
    Set the process-global icache.
    """
    return _object_cache


# TODO Use positional-only args for for with_object_cache
# once we drop Python 3.7 support.


def with_object_cache(cache_key, factory, *args, **kwargs):
    """
    Use value from cache or, if not present, call factory(*args, **kwargs) and cache result.
    """
    cache = get_object_cache()
    # If no cache configured, generate and return value.
    if cache is NO_CACHE:
        return factory(*args, **kwargs)
    # Return cached value if found.
    value = cache.get(cache_key)
    if value is not None:
        return value
    # Generate value and offer it to the cache, with an associated cost.
    start_time = time.perf_counter()
    value = factory(*args, **kwargs)
    cost = time.perf_counter() - start_time
    cache.put(cache_key, value, cost)
    return value


class ObjectCache:
    def __init__(self, available_bytes_in_process):
        self.misses = 0
        self.hits = 0

        def miss(key):
            self.misses += 1
            if __debug__:
                logger.debug("Miss %r", key)

        def hit(key, value):
            self.hits += 1
            if __debug__:
                logger.debug("Hit %r", key)

        self._cache = cachey.Cache(available_bytes_in_process, 0, miss=miss, hit=hit)
        self._dask_context = DaskCache(self)

    @property
    def dask_context(self):
        """
        Within this context, get and store dask tasks with the object cache.
        """
        return self._dask_context

    @property
    def available_bytes(self):
        """
        Maximum size in bytes
        """
        return self._cache.available_bytes

    def get(self, key):
        """
        Get cache item.
        """
        value = self._cache.get(key)
        return value

    def put(self, key, value, cost, nbytes=None):
        """
        Put cache item.

        Parameters
        ----------
        key : uniquely identifies content
        value : object
            May be any Python object. For future-proofing, the object should be
            pickle-able, as an _external_ object cache will be added in the future.
        cost : float
            Time in seconds that this value cost to obtain.
        nbytes : bytesize, optional
            Computed (with best effort) if not provided.
        """
        if nbytes is None:
            nbytes = self._cache.get_nbytes(value)
        logger.debug("Store %r (cost=%.3f, nbytes=%d)", key, cost, nbytes)
        self._cache.put(key, value, cost, nbytes=nbytes)

    def discard(self, *keys):
        """
        Discard one or more items from the cache if present.
        """
        for key in keys:
            # Cachey has no API specifically for this, but we can do it ourselves
            # but modifying only public state.
            value = self._cache.data.pop(key, None)
            if value is not None:
                self._cache.total_bytes -= self._cache.nbytes.pop(key)

    def discard_dask(self, *keys):
        """
        Discard one or more dask tasks from the cache, if present.

        Internally, cached dask keys are prefixed to avoid collisions.
        That is why it has a method separate from discard().
        """
        self.discard(("dask", key) for key in keys)

    def clear(self):
        """
        Empty the cache.
        """
        return self._cache.clear()

    def __contains__(self, key):
        return key in self._cache


class DaskCache(Callback):
    """
    Adapted from dask.cache

    Changes:
    - Scope keys under prefix 'dask-' to avoid collisions with non-dask cache usage
    - Use a simpler cost computation: duration (in seconds)
    """

    def __init__(self, cache):
        self._nbytes = cachey.nbytes
        self.cache = cache
        self.starttimes = dict()

    def _start(self, dsk):
        "Patched as noted in comment below"
        self.durations = dict()
        for key in dsk:
            # Use 'get', not cache.data[key] as upstream does,
            # in order to trip 'hit' and 'miss' callbacks.
            cached_result = self.cache.get(("dask", *key))
            if cached_result is not None:
                dsk[key] = cached_result

    def _pretask(self, key, dsk, state):
        self.starttimes[key] = default_timer()

    def _posttask(self, key, value, dsk, state, id):
        duration = default_timer() - self.starttimes[key]
        deps = state["dependencies"][key]
        if deps:
            duration += max(self.durations.get(k, 0) for k in deps)
        self.durations[key] = duration
        nb = self._nbytes(value)
        self.cache.put(("dask", *key), value, cost=duration, nbytes=nb)

    def _finish(self, dsk, state, errored):
        for key in dsk:
            self.starttimes.pop(key, None)
            self.durations.pop(key, None)



from dataclasses import dataclass
from typing import Protocol

from fastapi import Request


@dataclass
class UserSessionState:
    """Data transfer class to communicate custom session state infromation."""

    user_name: str
    state: dict = None


class UsernamePasswordAuthenticator(Protocol):
    def authenticate(self, username: str, password: str) -> UserSessionState:
        pass


class Authenticator(Protocol):
    def authenticate(self, request: Request) -> UserSessionState:
        pass



import asyncio
import string
import sys
from datetime import timedelta

import awkward
import numpy
import pandas
import sparse
import xarray

from tiled.adapters.array import ArrayAdapter
from tiled.adapters.awkward import AwkwardAdapter
from tiled.adapters.dataframe import DataFrameAdapter
from tiled.adapters.mapping import MapAdapter
from tiled.adapters.sparse import COOAdapter
from tiled.adapters.xarray import DatasetAdapter

print("Generating large example data...", file=sys.stderr)
data = {
    "big_image": numpy.random.random((10_000, 10_000)),
    "small_image": numpy.random.random((300, 300)),
    "medium_image": numpy.random.random((1000, 1000)),
    "tiny_image": numpy.random.random((50, 50)),
    "tiny_cube": numpy.random.random((50, 50, 50)),
    "tiny_hypercube": numpy.random.random((50, 50, 50, 50, 50)),
    "high_entropy": numpy.random.random((100, 100)),
    "low_entropy": numpy.ones((100, 100)),
    "short_column": numpy.random.random(100),
    "tiny_column": numpy.random.random(10),
    "long_column": numpy.random.random(100_000),
}
temp = 15 + 8 * numpy.random.randn(2, 2, 3)
precip = 10 * numpy.random.rand(2, 2, 3)
lon = [[-99.83, -99.32], [-99.79, -99.23]]
lat = [[42.25, 42.21], [42.63, 42.59]]
sparse_arr = numpy.random.random((100, 100))
sparse_arr[sparse_arr < 0.9] = 0  # fill most of the array with zeros
awkward_arr = awkward.Array(
    [[{"x": 1.1, "y": [1]}, {"x": 2.2, "y": [1, 2]}], [], [{"x": 3.3, "y": [1, 2, 3]}]]
)

print("Done generating example data.", file=sys.stderr)

mapping = {
    "big_image": ArrayAdapter.from_array(data["big_image"]),
    "small_image": ArrayAdapter.from_array(data["small_image"]),
    "medium_image": ArrayAdapter.from_array(data["medium_image"]),
    "sparse_image": COOAdapter.from_coo(sparse.COO(sparse_arr)),
    "awkward_array": AwkwardAdapter.from_array(awkward_arr),
    "tiny_image": ArrayAdapter.from_array(data["tiny_image"]),
    "tiny_cube": ArrayAdapter.from_array(data["tiny_cube"]),
    "tiny_hypercube": ArrayAdapter.from_array(data["tiny_hypercube"]),
    "short_table": DataFrameAdapter.from_pandas(
        pandas.DataFrame(
            {
                "A": data["short_column"],
                "B": 2 * data["short_column"],
                "C": 3 * data["short_column"],
            },
            index=pandas.Index(numpy.arange(len(data["short_column"])), name="index"),
        ),
        npartitions=1,
        metadata={"animal": "dog", "color": "red"},
    ),
    "long_table": DataFrameAdapter.from_pandas(
        pandas.DataFrame(
            {
                "A": data["long_column"],
                "B": 2 * data["long_column"],
                "C": 3 * data["long_column"],
            },
            index=pandas.Index(numpy.arange(len(data["long_column"])), name="index"),
        ),
        npartitions=5,
        metadata={"animal": "dog", "color": "green"},
    ),
    "wide_table": DataFrameAdapter.from_pandas(
        pandas.DataFrame(
            {
                letter: i * data["tiny_column"]
                for i, letter in enumerate(string.ascii_uppercase, start=1)
            },
            index=pandas.Index(numpy.arange(len(data["tiny_column"])), name="index"),
        ),
        npartitions=1,
        metadata={"animal": "dog", "color": "red"},
    ),
    "structured_data": MapAdapter(
        {
            "pets": ArrayAdapter.from_array(
                numpy.array(
                    [("Rex", 9, 81.0), ("Fido", 3, 27.0)],
                    dtype=[("name", "U10"), ("age", "i4"), ("weight", "f4")],
                )
            ),
            "xarray_dataset": DatasetAdapter.from_dataset(
                xarray.Dataset(
                    {
                        "temperature": (["x", "y", "time"], temp),
                        "precipitation": (["x", "y", "time"], precip),
                    },
                    coords={
                        "lon": (["x", "y"], lon),
                        "lat": (["x", "y"], lat),
                        "time": pandas.date_range("2014-09-06", periods=3),
                    },
                ),
            ),
        },
        metadata={"animal": "cat", "color": "green"},
    ),
    "flat_array": ArrayAdapter.from_array(numpy.random.random(100)),
    "low_entropy": ArrayAdapter.from_array(data["low_entropy"]),
    "high_entropy": ArrayAdapter.from_array(data["high_entropy"]),
    # Below, an asynchronous task modifies this value over time.
    "dynamic": ArrayAdapter.from_array(numpy.zeros((3, 3))),
}
# The entries aren't actually dynamic, but set entries_stale_after
# to demonstrate cache expiry.
tree = MapAdapter(mapping, entries_stale_after=timedelta(seconds=10))


async def increment_dynamic():
    """
    Change the value of the 'dynamic' node every 3 seconds.
    """
    fill_value = 0
    while True:
        fill_value += 1
        mapping["dynamic"] = ArrayAdapter.from_array(fill_value * numpy.ones((3, 3)))
        await asyncio.sleep(3)


# The server will run this on its event loop. We cannot start it *now* because
# there is not yet a running event loop.
tree.background_tasks.append(increment_dynamic)



"""Reorganize 'structure' and rename 'dataframe' to 'table'.

Revision ID: 83889e049ddc
Revises: 6825c778aa3c
Create Date: 2023-08-04 06:38:48.775874

"""
import base64

import pyarrow
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects.postgresql import JSONB

from tiled.serialization.table import deserialize_arrow
from tiled.structures.core import StructureFamily
from tiled.structures.table import B64_ENCODED_PREFIX

# revision identifiers, used by Alembic.
revision = "83889e049ddc"
down_revision = "6825c778aa3c"
branch_labels = None
depends_on = None


# Use JSON with SQLite and JSONB with PostgreSQL.
JSONVariant = sa.JSON().with_variant(JSONB(), "postgresql")


def upgrade():
    connection = op.get_bind()

    if connection.engine.dialect.name == "postgresql":
        # This change must be committed befor ethe new 'table' enum value can be used.
        with op.get_context().autocommit_block():
            op.execute(
                sa.text(
                    "ALTER TYPE structurefamily ADD VALUE IF NOT EXISTS 'table' AFTER 'dataframe'"
                )
            )
    # Rename "dataframe" to "table".
    nodes = sa.Table(
        "nodes",
        sa.MetaData(),
        sa.Column("id", sa.Integer),
        sa.Column("structure_family", sa.Unicode(32)),
    )
    # Use a raw text query to work around type fussiness in postgres.
    op.execute(
        sa.text(
            "UPDATE nodes SET structure_family='table' WHERE nodes.structure_family = 'dataframe'"
        )
    )

    data_sources = sa.Table(
        "data_sources",
        sa.MetaData(),
        sa.Column("id", sa.Integer),
        sa.Column("node_id", sa.Integer),
        sa.Column("structure", JSONVariant),
    )
    joined = data_sources.join(nodes, data_sources.c.node_id == nodes.c.id)
    results = connection.execute(
        sa.select(
            data_sources.c.id,
            data_sources.c.structure,
            nodes.c.structure_family,
        ).select_from(joined)
    ).fetchall()
    for id_, structure, structure_family in results:
        if structure_family == StructureFamily.array:
            # Consolidate "macro" and "micro".
            new_structure = {}
            new_structure["shape"] = structure["macro"]["shape"]
            new_structure["dims"] = structure["macro"]["dims"]
            new_structure["chunks"] = structure["macro"]["chunks"]
            new_structure["resizable"] = structure["macro"]["resizable"]
            new_structure["data_type"] = structure["micro"]
        elif structure_family == StructureFamily.table:
            # Consolidate "macro" and "micro".
            new_structure = {}
            new_structure["columns"] = structure["macro"]["columns"]
            new_structure["npartitions"] = structure["macro"]["npartitions"]
            new_structure["resizable"] = structure["macro"]["resizable"]
            # Re-encode the Arrow schema.
            meta_bytes = structure["micro"]["meta"]
            meta = deserialize_arrow(base64.b64decode(meta_bytes))
            schema_bytes = pyarrow.Table.from_pandas(meta).schema.serialize()
            schema_b64 = base64.b64encode(schema_bytes).decode("utf-8")
            data_uri = B64_ENCODED_PREFIX + schema_b64
            new_structure["arrow_schema"] = data_uri
        else:
            continue
        connection.execute(
            data_sources.update()
            .where(data_sources.c.id == id_)
            .values(structure=new_structure)
        )


def downgrade():
    # This _could_ be implemented but we will wait for a need since we are
    # still in alpha releases.
    raise NotImplementedError



import os
import sys
import tempfile
from pathlib import Path

import asyncpg
import pytest
import pytest_asyncio
from sqlalchemy.ext.asyncio import create_async_engine

from .. import profiles
from ..catalog import from_uri, in_memory
from ..server.settings import get_settings
from .utils import enter_password as utils_enter_password
from .utils import temp_postgres


@pytest.fixture(autouse=True)
def reset_settings():
    """
    Reset the FastAPI Settings.

    Fast API uses a global singleton for Settings.  It is difficult to get
    around this, so our best option is to reset it.
    """
    get_settings.cache_clear()
    yield


@pytest.fixture(autouse=True)
def set_tiled_cache_dir():
    """
    Use a tmpdir instead of ~/.cache/tiled/tokens
    """
    # Do not use tempdir pytest fixture because it would use the same tmpdir
    # as the one used by the test, and mix the files up.
    # Windows will not remove the directory while the http_response_cache.db
    # is still open. It is closed by transport shutdown, but not all tests
    # correctly shut down the transport. This is probably related to the
    # thread-leaking issue.
    # This option was added to TemporaryDirectory in Python 3.10
    kwargs = {}
    if sys.platform.startswith("win") and sys.version_info >= (3, 10):
        kwargs["ignore_cleanup_errors"] = True
    with tempfile.TemporaryDirectory(**kwargs) as tmpdir:
        os.environ["TILED_CACHE_DIR"] = str(tmpdir)
        yield
        del os.environ["TILED_CACHE_DIR"]


@pytest.fixture
def tmp_profiles_dir():
    """
    Use a tmpdir instead of ~/.config/tiled/profiles
    """
    # Do not use tempdir pytest fixture because it would use the same tmpdir
    # as the one used by the test, and mix the files up.
    with tempfile.TemporaryDirectory() as tmpdir:
        original = profiles.paths
        profiles.paths.clear()
        profiles.paths.extend([Path(tmpdir)])
        profiles.load_profiles.cache_clear()
        yield
        profiles.paths.clear()
        profiles.paths.extend(original)


@pytest.fixture
def enter_password():
    """
    DEPRECATED: Use the normal (non-fixture) context manager in .utils.
    """
    return utils_enter_password


@pytest.fixture(scope="module")
def tmpdir_module(request, tmpdir_factory):
    """A tmpdir fixture for the module scope. Persists throughout the module."""
    # Source: https://stackoverflow.com/a/31889843
    return tmpdir_factory.mktemp(request.module.__name__)


# Use this with pytest --log-cli-level=25 option.
if os.getenv("TILED_DEBUG_LEAKED_THREADS"):
    import logging
    import threading
    import time

    def poll_enumerate():
        logger = logging.getLogger(__name__)
        msg_level = int(logging.INFO + logging.WARNING) // 2
        while True:
            time.sleep(1)
            logger.log(msg_level, "THREAD COUNT = %d", len(threading.enumerate()))

    thread = threading.Thread(target=poll_enumerate, daemon=True)
    thread.start()


# To test with postgres, start a container like:
#
# docker run --name tiled-test-postgres -p 5432:5432 -e POSTGRES_PASSWORD=secret -d docker.io/postgres:16
# and set this env var like:
#
# TILED_TEST_POSTGRESQL_URI=postgresql+asyncpg://postgres:secret@localhost:5432

TILED_TEST_POSTGRESQL_URI = os.getenv("TILED_TEST_POSTGRESQL_URI")


@pytest_asyncio.fixture
async def postgresql_adapter(request, tmpdir):
    """
    Adapter instance

    Note that startup() and shutdown() are not called, and must be run
    either manually (as in the fixture 'a') or via the app (as in the fixture 'client').
    """
    if not TILED_TEST_POSTGRESQL_URI:
        raise pytest.skip("No TILED_TEST_POSTGRESQL_URI configured")
    # Create temporary database.
    async with temp_postgres(TILED_TEST_POSTGRESQL_URI) as uri_with_database_name:
        # Build an adapter on it, and initialize the database.
        adapter = from_uri(
            uri_with_database_name,
            writable_storage=str(tmpdir),
            init_if_not_exists=True,
        )
        yield adapter


@pytest_asyncio.fixture
async def postgresql_with_example_data_adapter(request, tmpdir):
    """
    Adapter instance

    Note that startup() and shutdown() are not called, and must be run
    either manually (as in the fixture 'a') or via the app (as in the fixture 'client').
    """
    if not TILED_TEST_POSTGRESQL_URI:
        raise pytest.skip("No TILED_TEST_POSTGRESQL_URI configured")
    DATABASE_NAME = "tiled-example-data"
    uri = TILED_TEST_POSTGRESQL_URI
    if uri.endswith("/"):
        uri = uri[:-1]
    uri_with_database_name = f"{uri}/{DATABASE_NAME}"
    engine = create_async_engine(uri_with_database_name)
    try:
        async with engine.connect():
            pass
    except asyncpg.exceptions.InvalidCatalogNameError:
        raise pytest.skip(
            f"PostgreSQL instance contains no database named {DATABASE_NAME!r}"
        )
    adapter = from_uri(
        uri_with_database_name,
        writable_storage=str(tmpdir),
    )
    yield adapter


@pytest_asyncio.fixture
async def sqlite_adapter(request, tmpdir):
    """
    Adapter instance

    Note that startup() and shutdown() are not called, and must be run
    either manually (as in the fixture 'a') or via the app (as in the fixture 'client').
    """
    yield in_memory(writable_storage=str(tmpdir))


@pytest_asyncio.fixture
async def sqlite_with_example_data_adapter(request, tmpdir):
    """
    Adapter instance

    Note that startup() and shutdown() are not called, and must be run
    either manually (as in the fixture 'a') or via the app (as in the fixture 'client').
    """
    SQLITE_DATABASE_PATH = Path("./tiled_test_db_sqlite.db")
    if not SQLITE_DATABASE_PATH.is_file():
        raise pytest.skip(f"Could not find {SQLITE_DATABASE_PATH}")
    adapter = from_uri(
        f"sqlite+aiosqlite:///{SQLITE_DATABASE_PATH}",
        writable_storage=str(tmpdir),
    )
    yield adapter


@pytest.fixture(params=["sqlite_adapter", "postgresql_adapter"])
def adapter(request):
    """
    Adapter instance

    Note that startup() and shutdown() are not called, and must be run
    either manually (as in the fixture 'a') or via the app (as in the fixture 'client').
    """
    yield request.getfixturevalue(request.param)


@pytest.fixture(
    params=["sqlite_with_example_data_adapter", "postgresql_with_example_data_adapter"]
)
def example_data_adapter(request):
    """
    Adapter instance

    Note that startup() and shutdown() are not called, and must be run
    either manually (as in the fixture 'a') or via the app (as in the fixture 'client').
    """
    yield request.getfixturevalue(request.param)



from pathlib import Path

import numpy
import pandas
import tifffile

from ..utils import modules_available

TIFF_PATHS = [
    ("a.tif",),
    ("b.tif",),
    ("c.tif",),
    ("more", "A0001.tif"),
    ("more", "A0002.tif"),
    ("more", "A0003.tif"),
    ("more", "B0001.tif"),
    ("more", "B0002.tif"),
    ("more", "even_more", "e.tif"),
    ("more", "even_more", "f.tif"),
]

# TODO Generate more interesting images.
data = numpy.ones((100, 100))

df1 = pandas.DataFrame({"A": [1, 2, 3], "B": [4, 5, 6]})
df2 = pandas.DataFrame({"C": [10, 20, 30], "D": [40, 50, 60]})


def generate_files(root_path):
    for path in TIFF_PATHS:
        full_path = Path(root_path, *path)
        full_path.parent.mkdir(parents=True, exist_ok=True)
        tifffile.imwrite(str(full_path), data)
    with pandas.ExcelWriter(Path(root_path, "tables.xlsx")) as writer:
        df1.to_excel(writer, sheet_name="Sheet 1", index=False)
        df2.to_excel(writer, sheet_name="Sheet 2", index=False)
    df1.to_csv(Path(root_path, "another_table.csv"))


if __name__ == "__main__":
    import sys

    if not modules_available("tifffile", "openpyxl"):
        print(
            """This example requires tifffile and openpyxl, which can be installed by

pip install tifffile openpyxl
""",
            file=sys.stderr,
        )
    generate_files(*sys.argv[1:])



from .queries import KeysFilter
from .scopes import SCOPES
from .utils import Sentinel, SpecialUsers, import_object

ALL_ACCESS = Sentinel("ALL_ACCESS")
ALL_SCOPES = set(SCOPES)
PUBLIC_SCOPES = {"read:metadata", "read:data"}
NO_ACCESS = Sentinel("NO_ACCESS")


class DummyAccessPolicy:
    "Impose no access restrictions."

    def allowed_scopes(self, node, principal):
        return ALL_SCOPES

    def filters(self, node, principal, scopes):
        return []


class SimpleAccessPolicy:
    """
    A mapping of user names to lists of entries they have access to.

    This simple policy does not provide fine-grained control of scopes.
    Any restriction on scopes is applied the same to all users, except
    for an optional list of 'admins'.

    This is used in the test suite; it may be suitable for very simple
    deployments.

    >>> SimpleAccessPolicy({"alice": ["A", "B"], "bob": ["B"]}, provider="toy")
    """

    ALL = ALL_ACCESS

    def __init__(
        self, access_lists, *, provider, scopes=None, public=None, admins=None
    ):
        self.access_lists = {}
        self.provider = provider
        self.scopes = scopes if (scopes is not None) else ALL_SCOPES
        self.public = set(public or [])
        self.admins = set(admins or [])
        for key, value in access_lists.items():
            if isinstance(value, str):
                value = import_object(value)
            self.access_lists[key] = value

    def _get_id(self, principal):
        # Get the id (i.e. username) of this Principal for the
        # associated authentication provider.
        for identity in principal.identities:
            if identity.provider == self.provider:
                id = identity.id
                break
        else:
            raise ValueError(
                f"Principcal {principal} has no identity from provider {self.provider}. "
                f"Its identities are: {principal.identities}"
            )
        return id

    def allowed_scopes(self, node, principal):
        # If this is being called, filter_access has let us get this far.
        if principal is SpecialUsers.public:
            allowed = PUBLIC_SCOPES
        elif self._get_id(principal) in self.admins:
            allowed = ALL_SCOPES
        # The simple policy does not provide for different Principals to
        # have different scopes on different Nodes. If the Principal has access,
        # they have the same hard-coded access everywhere.
        else:
            allowed = self.scopes
        return allowed

    def filters(self, node, principal, scopes):
        queries = []
        if principal is SpecialUsers.public:
            queries.append(KeysFilter(self.public))
        else:
            id = self._get_id(principal)
            if id in self.admins:
                return queries
            if not scopes.issubset(self.scopes):
                return NO_ACCESS
            access_list = self.access_lists.get(id, [])
            if not ((principal is SpecialUsers.admin) or (access_list == self.ALL)):
                try:
                    allowed = set(access_list or [])
                except TypeError:
                    # Provide rich debugging info because we have encountered a confusing
                    # bug here in a previous implementation.
                    raise TypeError(
                        f"Unexpected access_list {access_list} of type {type(access_list)}. "
                        f"Expected iterable or {self.ALL}, instance of {type(self.ALL)}."
                    )
                queries.append(KeysFilter(allowed))
        return queries



from ..adapters.mapping import MapAdapter
from ..client import Context, from_context
from ..server.app import build_app
from ..structures.core import Spec


def test_specs():
    tree = MapAdapter({}, specs=[Spec("spec_test")])
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)
    assert client.specs == [Spec("spec_test")]


def test_specs_give_as_str():
    tree = MapAdapter({}, specs=["spec_test"])
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)
    assert client.specs == [Spec("spec_test")]


def test_specs_with_version():
    tree = MapAdapter({}, specs=[Spec("spec_test", version="1.1")])
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)
    assert client.specs == [Spec("spec_test", version="1.1")]



from ..utils import modules_available


def register_builtin_serializers():
    """
    Register built-in serializers for each structure if its associated module(s) are installed.
    """
    # Each submodule in ..serialization registers serializers on import.
    # Some are conditional on the availability of particular I/O libraries.
    from ..serialization import container as _container  # noqa: F401

    del _container
    if modules_available("numpy", "dask.array"):
        from ..serialization import array as _array  # noqa: F401

        del _array
    if modules_available("awkward"):
        from ..serialization import awkward as _awkward  # noqa: F401

        del _awkward
    if modules_available("pandas", "pyarrow", "dask.dataframe"):
        from ..serialization import table as _table  # noqa: F401

        del _table
    if modules_available("sparse"):
        from ..serialization import sparse as _sparse  # noqa: F401

        del _sparse
    if modules_available("xarray"):
        from ..serialization import xarray as _xarray  # noqa: F401

        del _xarray



import dask.dataframe
import pandas

from ..adapters.mapping import MapAdapter
from ..server.object_cache import NO_CACHE, get_object_cache, with_object_cache
from .dataframe import DataFrameAdapter


class ExcelAdapter(MapAdapter):
    @classmethod
    def from_file(cls, file, **kwargs):
        """
        Read the sheets in an Excel file.

        This maps the Excel file, which may contain one of more spreadsheets,
        onto a tree of tabular structures.

        Examples
        --------

        Given a file object

        >>> file = open("path/to/excel_file.xlsx")
        >>> ExcelAdapter.from_file(file)

        Given a pandas.ExcelFile object

        >>> import pandas
        >>> filepath = "path/to/excel_file.xlsx"
        >>> ef = pandas.ExcelFile(filepath)
        >>> ExcelAdapter.from_file(ef)
        """
        if isinstance(file, pandas.ExcelFile):
            excel_file = file
        else:
            excel_file = pandas.ExcelFile(file)
        # If an instance has previously been created using the same parameters,
        # then we are here because the caller wants a *fresh* view on this data.
        # Therefore, we should clear any cached data.
        cache = get_object_cache()
        mapping = {}
        for sheet_name in excel_file.sheet_names:
            cache_key = (cls.__module__, cls.__qualname__, file, sheet_name)
            ddf = dask.dataframe.from_pandas(
                with_object_cache(cache_key, excel_file.parse, sheet_name),
                npartitions=1,  # TODO Be smarter about this.
            )
            if cache is not NO_CACHE:
                cache.discard(cache_key)  # parsed sheet content
                cache.discard_dask(ddf.__dask_keys__())  # dask tasks
            mapping[sheet_name] = DataFrameAdapter.from_dask_dataframe(ddf)
        return cls(mapping, **kwargs)

    @classmethod
    def from_uri(cls, data_uri, **kwargs):
        """
        Read the sheets in an Excel file.

        This maps the Excel file, which may contain one of more spreadsheets,
        onto a tree of tabular structures.

        Examples
        --------

        Given a file path

        >>> ExcelAdapter.from_file("path/to/excel_file.xlsx")
        """
        file = pandas.ExcelFile(data_uri)
        return cls.from_file(file)



"""Initialize

Revision ID: 481830dd6c11
Revises:
Create Date: 2022-01-13 11:26:35.432786

"""
import uuid

from alembic import op
from sqlalchemy import (
    Boolean,
    Column,
    DateTime,
    Enum,
    ForeignKey,
    Integer,
    LargeBinary,
    Unicode,
)
from sqlalchemy.sql import func

from tiled.authn_database.orm import UUID, JSONList, PrincipalType

# revision identifiers, used by Alembic.
revision = "481830dd6c11"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "principals",
        Column("time_created", DateTime(timezone=False), server_default=func.now()),
        Column("time_updated", DateTime(timezone=False), onupdate=func.now()),
        Column("id", Integer, primary_key=True, index=True, autoincrement=True),
        Column(
            "uuid",
            UUID,
            index=True,
            nullable=False,
            default=lambda: uuid.uuid4(),
        ),
        Column("type", Enum(PrincipalType), nullable=False),
    )
    op.create_table(
        "identities",
        Column("time_created", DateTime(timezone=False), server_default=func.now()),
        Column("time_updated", DateTime(timezone=False), onupdate=func.now()),
        Column("id", Unicode(255), primary_key=True, nullable=False),
        Column("provider", Unicode(255), primary_key=True, nullable=False),
        Column("principal_id", Integer, ForeignKey("principals.id")),
        Column("latest_login", DateTime(timezone=False), nullable=True),
    )
    op.create_table(
        "roles",
        Column("time_created", DateTime(timezone=False), server_default=func.now()),
        Column("time_updated", DateTime(timezone=False), onupdate=func.now()),
        Column("id", Integer, primary_key=True, index=True, autoincrement=True),
        Column("name", Unicode(255), index=True, unique=True),
        Column("description", Unicode(1023), nullable=True),
        Column("scopes", JSONList(511), nullable=False),
    )
    op.create_table(
        "principal_role_association",
        Column("principal_id", Integer, ForeignKey("principals.id"), primary_key=True),
        Column("role_id", Integer, ForeignKey("roles.id"), primary_key=True),
    )
    op.create_table(
        "api_keys",
        Column("time_created", DateTime(timezone=False), server_default=func.now()),
        Column("time_updated", DateTime(timezone=False), onupdate=func.now()),
        Column(
            "hashed_secret",
            LargeBinary(32),
            primary_key=True,
            index=True,
            nullable=False,
        ),
        Column(
            "first_eight",
            Unicode(8),
            primary_eky=True,
            index=True,
            nullable=False,
        ),
        Column("latest_activity", DateTime(timezone=False), nullable=True),
        Column("expiration_time", DateTime(timezone=False), nullable=True),
        Column("note", Unicode(1023), nullable=True),
        Column("principal_id", Integer, ForeignKey("principals.id"), nullable=False),
        Column("scopes", JSONList(511), nullable=False),
    )
    op.create_table(
        "sessions",
        Column("time_created", DateTime(timezone=False), server_default=func.now()),
        Column("time_updated", DateTime(timezone=False), onupdate=func.now()),
        Column("id", Integer, primary_key=True, index=True, autoincrement=True),
        Column(
            "uuid",
            UUID,
            index=True,
            nullable=False,
            default=lambda: uuid.uuid4(),
        ),
        Column("time_last_refreshed", DateTime(timezone=False), nullable=True),
        Column("refresh_count", Integer, nullable=False, default=0),
        Column("expiration_time", DateTime(timezone=False), nullable=False),
        Column("principal_id", Integer, ForeignKey("principals.id"), nullable=False),
        Column("revoked", Boolean, default=False, nullable=False),
    )


def downgrade():
    "Nothing to do because this is the initial schema."
    pass



from dask.base import normalize_object, normalize_token
from dask.base import tokenize as dask_tokenize


def tokenize(obj):
    # This just uses dask for now but may evolve to something custom in future.
    return dask_tokenize(obj)


# When dask does not know how to deterministically tokenize something, it falls
# back to normalize_object, which generates a random, nonderministic token
# (UUID4). Here, we teach it how to tokenize HDF5 datasets. This function will
# only be run if/when an h5py object is encountered.


@normalize_token.register_lazy("h5py")
def register_h5py():
    from pathlib import Path

    import h5py

    @normalize_token.register(h5py.Dataset)
    def normalize_dataset(dataset):
        path = Path(dataset.file.filename)
        if path.is_file() and path.is_absolute():
            return (dataset.file.filename, path.stat().st_mtime, dataset.name)
        # If we reach here, we have some HDF5 file with only a *relative* path or
        # one not backed by a file on disk at all. (It could be BytesIO.)
        return normalize_object(dataset)



# Write the benchmarking functions here.
# See "Writing benchmarks" in the asv docs for more information.

import numpy

from tiled.adapters.array import ArrayAdapter
from tiled.adapters.mapping import MapAdapter
from tiled.client import Context, from_context
from tiled.server.app import build_app


class TimeSuite:
    def setup(self):
        tree = MapAdapter({"x": ArrayAdapter.from_array(numpy.ones((100, 100)))})
        app = build_app(tree)
        self.context = Context.from_app(app)
        self.client = from_context(self.context)

    def teardown(self):
        self.context.close()

    def time_list_tree(self):
        list(self.client)

    def time_lookup(self):
        self.client["x"]

    def time_lookup_and_read(self):
        self.client["x"].read()



"Adapted from https://github.com/sqlalchemy/sqlalchemy/wiki/Query-Plan-SQL-construct"
import contextlib
import os

from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql.expression import ClauseElement, Executable

EXPLAIN_SQL = bool(int(os.getenv("TILED_EXPLAIN_SQL", "0") or "0"))


class _explain(Executable, ClauseElement):
    inherit_cache = False

    def __init__(self, stmt, analyze=False):
        self.statement = stmt
        self.analyze = analyze


@compiles(_explain, "postgresql")
def pg_explain(element, compiler, **kw):
    text = "EXPLAIN "
    if element.analyze:
        text += "ANALYZE "
    text += compiler.process(element.statement, **kw)

    return text


@compiles(_explain, "sqlite")
def sqlite_explain(element, compiler, **kw):
    text = "EXPLAIN QUERY PLAN "
    if element.analyze:
        text += "EXPLAIN QUERY PLAN "
    text += compiler.process(element.statement, **kw)

    return text


_query_explanation_callbacks = []


@contextlib.contextmanager
def record_explanations():
    explanations = []

    def capture(e):
        explanations.append(e)

    _query_explanation_callbacks.append(capture)
    yield explanations
    _query_explanation_callbacks.remove(capture)


class ExplainAsyncSession(AsyncSession):
    """
    Extend AsyncSession to explain and then query.

    If EXPLAIN_SQL is off and there are now callbacks, just fall back
    to normal AsyncSession. For performance reasons, we only query
    for the explanation if it will be used.

    Use 'explain' paramete (default: None) to force explain on
    or off (True, False).

    1. Explain the query.
    2. Pass the explanation to the callback(s).
    3. Execute normally.
    """

    async def execute(self, statement, *args, explain=None, **kwargs):
        if explain is None:
            explain = bool(EXPLAIN_SQL or _query_explanation_callbacks)
        if explain:
            explanation = (
                await super().execute(_explain(statement), *args, **kwargs)
            ).all()
            for callback in _query_explanation_callbacks:
                callback(explanation)
        return await super().execute(statement, *args, **kwargs)



from ..client import Context
from ..server.app import build_app_from_config

strict_config = {
    "authentication": {
        "allow_anonymous_access": True,
    },
    "trees": [
        {
            "tree": "tiled.examples.generated_minimal:tree",
            "path": "/",
        },
    ],
}

permissive_config = strict_config.copy()
permissive_config["allow_origins"] = ["https://example.com"]
headers = {
    "Access-Control-Request-Method": "GET",
    "Origin": "https://example.com",
}


def test_cors_enforcement():
    with Context.from_app(build_app_from_config(strict_config)) as context:
        request = context.http_client.build_request("OPTIONS", "/", headers=headers)
        response = context.http_client.send(request)
        assert response.status_code == 400


def test_allow_origins():
    with Context.from_app(build_app_from_config(permissive_config)) as context:
        request = context.http_client.build_request("OPTIONS", "/", headers=headers)
        response = context.http_client.send(request)
        assert response.status_code == 200



import collections
import dataclasses
import logging
import mimetypes
import re
from pathlib import Path

import anyio
import watchfiles

from ..server.schemas import Asset, DataSource, Management
from ..structures.core import StructureFamily
from ..utils import import_object
from .adapter import Collision
from .mimetypes import DEFAULT_ADAPTERS_BY_MIMETYPE, DEFAULT_MIMETYPES_BY_FILE_EXT
from .utils import ensure_uri

logger = logging.getLogger(__name__)


def strip_suffixes(filename):
    """
    For use with key_from_filename parameter.

    This gives the 'base' as the key.

    >>> strip_suffixes("a.tif")
    "a"

    >>> strip_suffixes("thing.tar.gz")
    "thing"
    """
    path = Path(filename)
    # You would think there would be a method for this, but there is not.
    if len(path.suffixes):
        return str(path)[: -sum([len(s) for s in path.suffixes])]
    else:
        return filename


def identity(filename):
    """
    For use with key_from_filename parameter.

    This give the full filename (with suffixes) as the key.
    """
    return filename


def default_filter(path):
    "By default, ignore only hidden files."
    return not path.name.startswith(".")


def resolve_mimetype(path, mimetypes_by_file_ext, mimetype_detection_hook=None):
    """
    Given a filepath (file or directory) detect the mimetype.

    If no mimetype could be resolved, return None.
    """
    # First, try to infer the mimetype from the file extension.
    # For compound suffixes like '.u1.strict_disabled.avif' (a real example)
    # consider in order:
    # '.u1.strict_disabled.avif'
    # '.strict_disabled.avif'
    # '.avif'
    for i in range(len(path.suffixes)):
        ext = "".join(path.suffixes[i:])  # e.g. ".h5" or ".tar.gz"
        if ext in mimetypes_by_file_ext:
            mimetype = mimetypes_by_file_ext[ext]
            break
    else:
        # Use the Python's built-in facility for guessing mimetype
        # from file extension. This loads data about mimetypes from
        # the operating system the first time it is used.
        mimetype, _ = mimetypes.guess_type(str(path))
    # Finally, user-specified function has the opportunity to
    # look at more than just the file extension. This gets access to the full
    # path, so it can consider the file name and even open the file. It is also
    # passed the mimetype determined above, or None if no match was found.
    if mimetype_detection_hook is not None:
        mimetype = mimetype_detection_hook(path, mimetype)
    return mimetype


@dataclasses.dataclass(frozen=True)
class Settings:
    adapters_by_mimetype: dict
    mimetypes_by_file_ext: dict
    mimetype_detection_hook: callable
    key_from_filename: callable
    filter: callable

    @classmethod
    def init(
        cls,
        adapters_by_mimetype=None,
        mimetypes_by_file_ext=None,
        mimetype_detection_hook=None,
        key_from_filename=None,
        filter=None,
    ):
        # If parameters come from a configuration file, they are given
        # are given as importable strings, like "package.module:Reader".
        adapters_by_mimetype = adapters_by_mimetype or {}
        for key, value in list((adapters_by_mimetype).items()):
            if isinstance(value, str):
                adapters_by_mimetype[key] = import_object(value)
        merged_adapters_by_mimetype = collections.ChainMap(
            adapters_by_mimetype, DEFAULT_ADAPTERS_BY_MIMETYPE
        )
        if isinstance(key_from_filename, str):
            key_from_filename = import_object(key_from_filename)
        elif key_from_filename is None:
            key_from_filename = strip_suffixes
        if mimetype_detection_hook is not None:
            mimetype_detection_hook = import_object(mimetype_detection_hook)
        merged_mimetypes_by_file_ext = collections.ChainMap(
            mimetypes_by_file_ext or {}, DEFAULT_MIMETYPES_BY_FILE_EXT
        )
        if filter is None:
            filter = default_filter
        if isinstance(filter, str):
            filter = import_object(filter)
        return cls(
            adapters_by_mimetype=merged_adapters_by_mimetype,
            mimetypes_by_file_ext=merged_mimetypes_by_file_ext,
            mimetype_detection_hook=mimetype_detection_hook,
            key_from_filename=key_from_filename,
            filter=filter,
        )


async def register(
    catalog,
    path,
    prefix="/",
    walkers=None,
    adapters_by_mimetype=None,
    mimetypes_by_file_ext=None,
    mimetype_detection_hook=None,
    key_from_filename=None,
    filter=None,
    overwrite=True,
):
    "Register a file or directory (recursively)."
    settings = Settings.init(
        adapters_by_mimetype=adapters_by_mimetype,
        mimetypes_by_file_ext=mimetypes_by_file_ext,
        mimetype_detection_hook=mimetype_detection_hook,
        key_from_filename=key_from_filename,
        filter=filter,
    )
    path = Path(path)
    parsed_walkers = []
    for walker in walkers or []:
        parsed_walkers.append(import_object(walker))
    parsed_walkers.extend(DEFAULT_WALKERS)
    prefix_parts = [segment for segment in prefix.split("/") if segment]
    for segment in prefix_parts:
        child_catalog = await catalog.lookup_adapter([segment])
        if child_catalog is None:
            key = key_from_filename(segment)
            await create_node_safe(
                catalog,
                structure_family=StructureFamily.container,
                metadata={},
                specs=[],
                key=key,
            )
            child_catalog = await catalog.lookup_adapter([segment])
        catalog = child_catalog
    if path.is_dir():
        # Recursively enter the directory and any subdirectories.
        if overwrite:
            logger.info(f"  Overwriting '/{'/'.join(prefix_parts)}'")
            await catalog.delete_tree()
        await _walk(
            catalog,
            Path(path),
            parsed_walkers,
            settings=settings,
        )
    else:
        await register_single_item(
            catalog,
            path,
            is_directory=False,
            settings=settings,
        )


async def _walk(
    catalog,
    path,
    walkers,
    settings,
):
    "This is the recursive inner loop of walk."
    files = []
    directories = []
    logger.info("  Walking '%s'", path)
    for item in path.iterdir():
        if not settings.filter(item):
            continue
        if item.is_dir():
            directories.append(item)
        else:
            files.append(item)
    for walker in walkers:
        files, directories = await walker(
            catalog,
            path,
            files,
            directories,
            settings,
        )
    for directory in directories:
        key = settings.key_from_filename(directory.name)
        await create_node_safe(
            catalog,
            key=key,
            structure_family=StructureFamily.container,
            metadata={},
            specs=[],
        )
        child_catalog = await catalog.lookup_adapter([key])
        await _walk(
            child_catalog,
            directory,
            walkers,
            settings,
        )


async def one_node_per_item(
    catalog,
    path,
    files,
    directories,
    settings,
):
    "Process each file and directory as mapping to one logical 'node' in Tiled."
    unhandled_files = []
    unhandled_directories = []
    for file in files:
        result = await register_single_item(
            catalog,
            file,
            is_directory=False,
            settings=settings,
        )
        if not result:
            unhandled_files.append(file)
    for directory in directories:
        result = await register_single_item(
            catalog,
            directory,
            is_directory=True,
            settings=settings,
        )
        if not result:
            unhandled_directories.append(directory)
    return unhandled_files, unhandled_directories


async def register_single_item(
    catalog,
    item,
    is_directory,
    settings,
):
    "Register a single file or directory as a node."
    unhandled_items = []
    mimetype = resolve_mimetype(
        item, settings.mimetypes_by_file_ext, settings.mimetype_detection_hook
    )
    if mimetype is None:
        unhandled_items.append(item)
        if not is_directory:
            logger.info("    SKIPPED: Could not resolve mimetype for '%s'", item)
        return
    if mimetype not in settings.adapters_by_mimetype:
        logger.info(
            "    SKIPPED: Resolved mimetype '%s' but no adapter found for '%s'",
            mimetype,
            item,
        )
        unhandled_items.append(item)
        return
    adapter_factory = settings.adapters_by_mimetype[mimetype]
    logger.info("    Resolved mimetype '%s' with adapter for '%s'", mimetype, item)
    try:
        adapter = await anyio.to_thread.run_sync(adapter_factory, ensure_uri(item))
    except Exception:
        logger.exception("    SKIPPED: Error constructing adapter for '%s':", item)
        return
    key = settings.key_from_filename(item.name)
    return await create_node_safe(
        catalog,
        key=key,
        structure_family=adapter.structure_family,
        metadata=dict(adapter.metadata()),
        specs=adapter.specs,
        data_sources=[
            DataSource(
                mimetype=mimetype,
                structure=dict_or_none(adapter.structure()),
                parameters={},
                management=Management.external,
                assets=[
                    Asset(
                        data_uri=ensure_uri(item),
                        is_directory=is_directory,
                        parameter="data_uri",
                    )
                ],
            )
        ],
    )


# Matches filename with (optional) prefix characters followed by digits \d
# and then the file extension .tif or .tiff.
TIFF_SEQUENCE_STEM_PATTERN = re.compile(r"^(.*?)(\d+)\.(?:tif|tiff)$")
TIFF_SEQUENCE_EMPTY_NAME_ROOT = "_unnamed"


async def group_tiff_sequences(
    catalog,
    path,
    files,
    directories,
    settings,
):
    """
    Group files in the given directory into TIFF sequences.

    We are looking for any files:
    - with file extension .tif or .tiff
    - with file name ending in a number

    We group these into sorted groups and make one Node for each.
    A group may have one or more items.
    """
    unhandled_directories = directories
    unhandled_files = []
    sequences = collections.defaultdict(list)
    for file in files:
        if file.is_file():
            match = TIFF_SEQUENCE_STEM_PATTERN.match(file.name)
            if match:
                sequence_name, _sequence_number = match.groups()
                if sequence_name == "":
                    sequence_name = TIFF_SEQUENCE_EMPTY_NAME_ROOT
                sequences[sequence_name].append(file)
                continue
        unhandled_files.append(file)
    for name, sequence in sequences.items():
        await register_tiff_sequence(catalog, name, sorted(sequence), settings)
    return unhandled_files, unhandled_directories


TIFF_SEQ_MIMETYPE = "multipart/related;type=image/tiff"


async def register_tiff_sequence(catalog, name, sequence, settings):
    logger.info("    Grouped %d TIFFs into a sequence '%s'", len(sequence), name)
    adapter_class = settings.adapters_by_mimetype[TIFF_SEQ_MIMETYPE]
    key = settings.key_from_filename(name)
    try:
        adapter = adapter_class([ensure_uri(filepath) for filepath in sequence])
    except Exception:
        logger.exception("    SKIPPED: Error constructing adapter for '%s'", name)
        return
    await create_node_safe(
        catalog,
        key=key,
        structure_family=adapter.structure_family,
        metadata=dict(adapter.metadata()),
        specs=adapter.specs,
        data_sources=[
            DataSource(
                mimetype=TIFF_SEQ_MIMETYPE,
                structure=dict_or_none(adapter.structure()),
                parameters={},
                management=Management.external,
                assets=[
                    Asset(
                        data_uri=ensure_uri(item),
                        is_directory=False,
                        parameter="data_uris",
                        num=i,
                    )
                    for i, item in enumerate(sequence)
                ],
            )
        ],
    )


async def skip_all(
    catalog,
    path,
    files,
    directories,
    settings,
):
    """
    Skip all files and directories without processing them.

    This can be used to override the DEFAULT_WALKERS.
    """
    for item in files:
        logger.info("    SKIP ALL: Nothing yet handled file '%s'", item)
    return [], directories


DEFAULT_WALKERS = [group_tiff_sequences, one_node_per_item]


async def watch(
    catalog,
    path,
    prefix="/",
    walkers=None,
    adapters_by_mimetype=None,
    mimetypes_by_file_ext=None,
    mimetype_detection_hook=None,
    key_from_filename=None,
    filter=None,
    initial_walk_complete_event=None,
):
    settings = Settings.init(
        adapters_by_mimetype=adapters_by_mimetype,
        mimetypes_by_file_ext=mimetypes_by_file_ext,
        mimetype_detection_hook=mimetype_detection_hook,
        key_from_filename=key_from_filename,
        filter=filter,
    )
    if initial_walk_complete_event is None:
        initial_walk_complete_event = anyio.Event()
    ready_event = anyio.Event()
    stop_event = anyio.Event()
    async with anyio.create_task_group() as tg:
        # Begin listening for changes.
        tg.start_soon(
            _watch,
            ready_event,
            initial_walk_complete_event,
            stop_event,
            catalog,
            path,
            prefix,
            walkers,
            settings,
        )
        await ready_event.wait()
        # We have begun listening for changes.
        # Now do the initial walk.
        await register(
            catalog,
            path,
            prefix,
            walkers,
            adapters_by_mimetype=settings.adapters_by_mimetype,
            mimetypes_by_file_ext=settings.mimetypes_by_file_ext,
            mimetype_detection_hook=settings.mimetype_detection_hook,
            key_from_filename=settings.key_from_filename,
            filter=settings.filter,
        )
        # Signal that initial walk is complete.
        # Process any changes that were accumulated during the initial walk.
        await initial_walk_complete_event.set()


async def _watch(
    ready_event,
    initial_walk_complete_event,
    stop_event,
    catalog,
    path,
    prefix,
    walkers,
    settings,
):
    def watch_filter(change, path):
        return settings.filter(Path(path))

    await ready_event.set()
    backlog = []
    async for batch in watchfiles.awatch(
        path,
        watch_filter=watch_filter,
        yield_on_timeout=True,
        stop_event=stop_event,
        rust_timeout=1000,
    ):
        if (backlog is not None) and batch:
            logger.info(
                "Detected changes, waiting to process until initial walk completes"
            )
            backlog.extend(batch)
        if (backlog is not None) and initial_walk_complete_event.is_set():
            logger.info("Watching for changes in '%s'", path)
            if backlog:
                logger.info(
                    "Processing backlog of changes that occurred during initial walk..."
                )
                await process_changes(
                    batch,
                    catalog,
                    path,
                    prefix,
                    walkers,
                    settings,
                )
            backlog = None
        elif batch:
            # We are caught up. Process changes immediately.
            logger.info("Detected changes")
            await process_changes(
                batch,
                catalog,
                path,
                prefix,
                walkers,
                settings,
            )


async def process_changes(
    batch,
    catalog,
    path,
    prefix,
    walkers,
    settings,
):
    for change in batch:
        change_type, change_path = change
        logger.info("  %s '%s'", change_type.name, change_path)
        # TODO Be more selective.
        # We should be able to re-register only a select portion of the
        # full tree. For now, we ignore the change batch content and just
        # use the change as a timing signal to re-register the whole tree.
        await register(
            catalog,
            path,
            prefix,
            walkers,
            adapters_by_mimetype=settings.adapters_by_mimetype,
            mimetypes_by_file_ext=settings.mimetypes_by_file_ext,
            mimetype_detection_hook=settings.mimetype_detection_hook,
            key_from_filename=settings.key_from_filename,
            filter=settings.filter,
        )


async def create_node_safe(
    catalog,
    *args,
    key,
    **kwargs,
):
    "Call catalog.create_node(...) and if there is a collision remove the original."
    try:
        return await catalog.create_node(*args, key=key, **kwargs)
    except Collision as err:
        # To avoid ambiguity include _neither_ the original nor the new one.
        offender = await catalog.lookup_adapter([key])
        await offender.delete_tree()
        logger.warning(
            "   COLLISION: Multiple files would result in node at '%s'. Skipping all.",
            err.args[0],
        )


def dict_or_none(structure):
    if structure is None:
        return None
    return dataclasses.asdict(structure)



from functools import lru_cache
from typing import Optional

import pydantic
from fastapi import Depends, HTTPException, Query, Request, Security

from ..media_type_registration import (
    deserialization_registry as default_deserialization_registry,
)
from ..media_type_registration import (
    serialization_registry as default_serialization_registry,
)
from ..query_registration import query_registry as default_query_registry
from ..validation_registration import validation_registry as default_validation_registry
from .authentication import get_current_principal, get_session_state
from .core import NoEntry
from .utils import filter_for_access, record_timing


@lru_cache(1)
def get_query_registry():
    "This may be overridden via dependency_overrides."
    return default_query_registry


@lru_cache(1)
def get_deserialization_registry():
    "This may be overridden via dependency_overrides."
    return default_deserialization_registry


@lru_cache(1)
def get_serialization_registry():
    "This may be overridden via dependency_overrides."
    return default_serialization_registry


@lru_cache(1)
def get_validation_registry():
    "This may be overridden via dependency_overrides."
    return default_validation_registry


def get_root_tree():
    raise NotImplementedError(
        "This should be overridden via dependency_overrides. "
        "See tiled.server.app.build_app()."
    )


def SecureEntry(scopes):
    async def inner(
        path: str,
        request: Request,
        principal: str = Depends(get_current_principal),
        root_tree: pydantic.BaseSettings = Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
    ):
        """
        Obtain a node in the tree from its path.

        Walk down the path from the root tree, filtering each intermediate node by
        'read:metadata' and finally filtering by the specified scope.

        session_state is an optional dictionary passed in the session token
        """
        path_parts = [segment for segment in path.split("/") if segment]
        entry = root_tree

        # If the entry/adapter can take a session state, pass it in.
        # The entry/adapter may return itself or a different object.
        if hasattr(entry, "with_session_state") and session_state:
            entry = entry.with_session_state(session_state)
        try:
            # Traverse into sub-tree(s). This requires only 'read:metadata' scope.
            for i, segment in enumerate(path_parts):
                # add session state to entry
                entry = filter_for_access(
                    entry, principal, ["read:metadata"], request.state.metrics
                )
                # The new catalog adapter only has access control at top level for now.
                # It can jump directly to the node of interest.

                if hasattr(entry, "lookup_adapter"):
                    entry = await entry.lookup_adapter(path_parts[i:])
                    if entry is None:
                        raise NoEntry(path_parts)
                    break
                # Old-style dict-like interface
                else:
                    try:
                        entry = entry[segment]
                    except (KeyError, TypeError):
                        raise NoEntry(path_parts)
            # Now check that we have the requested scope on the final node.
            access_policy = getattr(entry, "access_policy", None)
            if access_policy is not None:
                with record_timing(request.state.metrics, "acl"):
                    allowed_scopes = entry.access_policy.allowed_scopes(
                        entry, principal
                    )
                    if not set(scopes).issubset(allowed_scopes):
                        if "read:metadata" not in allowed_scopes:
                            # If you can't read metadata, it does not exit for you.
                            raise NoEntry(path_parts)
                        else:
                            # You can see this, but you cannot perform the requested
                            # operation on it.
                            raise HTTPException(
                                status_code=403,
                                detail=(
                                    "Not enough permissions to perform this action on this node. "
                                    f"Requires scopes {scopes}. "
                                    f"Principal had scopes {list(allowed_scopes)} on this node."
                                ),
                            )
        except NoEntry:
            raise HTTPException(status_code=404, detail=f"No such entry: {path_parts}")
        return entry

    return Security(inner, scopes=scopes)


def block(
    # Ellipsis as the "default" tells FastAPI to make this parameter required.
    block: str = Query(..., pattern="^[0-9]*(,[0-9]+)*$"),
):
    "Specify and parse a block index parameter."
    if not block:
        return ()
    return tuple(map(int, block.split(",")))


def expected_shape(
    expected_shape: Optional[str] = Query(
        None, min_length=1, pattern="^[0-9]+(,[0-9]+)*$|^scalar$"
    ),
):
    "Specify and parse an expected_shape parameter."
    if expected_shape is None:
        return
    if expected_shape == "scalar":
        return ()
    return tuple(map(int, expected_shape.split(",")))


def slice_(
    slice: str = Query(None, pattern="^[-0-9,:]*$"),
):
    "Specify and parse a block index parameter."
    import numpy

    # IMPORTANT We are eval-ing a user-provider string here so we need to be
    # very careful about locking down what can be in it. The regex above
    # excludes any letters or operators, so it is not possible to execute
    # functions or expensive arithmetic.
    return tuple(
        [
            eval(f"numpy.s_[{dim!s}]", {"numpy": numpy})
            for dim in (slice or "").split(",")
            if dim
        ]
    )
